# -*- coding: utf-8 -*-
#  A modified version by Francois Gallard, from :
# Vahid Moosavi 2015 02 04 10:08 pm
# sevamoo@gmail.com
# Chair For Computer Aided Architectural Design, ETH  Zurich
# Future Cities Lab
# www.vahidmoosavi.com
#         Apache License
#                            Version 2.0, January 2004
#                         http://www.apache.org/licenses/
#
#    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
#
#    1. Definitions.
#
#       "License" shall mean the terms and conditions for use, reproduction,
#       and distribution as defined by Sections 1 through 9 of this document.
#
#       "Licensor" shall mean the copyright owner or entity authorized by
#       the copyright owner that is granting the License.
#
#       "Legal Entity" shall mean the union of the acting entity and all
#       other entities that control, are controlled by, or are under common
#       control with that entity. For the purposes of this definition,
#       "control" means (i) the power, direct or indirect, to cause the
#       direction or management of such entity, whether by contract or
#       otherwise, or (ii) ownership of fifty percent (50%) or more of the
#       outstanding shares, or (iii) beneficial ownership of such entity.
#
#       "You" (or "Your") shall mean an individual or Legal Entity
#       exercising permissions granted by this License.
#
#       "Source" form shall mean the preferred form for making modifications,
#       including but not limited to software source code, documentation
#       source, and configuration files.
#
#       "Object" form shall mean any form resulting from mechanical
#       transformation or translation of a Source form, including but
#       not limited to compiled object code, generated documentation,
#       and conversions to other media types.
#
#       "Work" shall mean the work of authorship, whether in Source or
#       Object form, made available under the License, as indicated by a
#       copyright notice that is included in or attached to the work
#       (an example is provided in the Appendix below).
#
#       "Derivative Works" shall mean any work, whether in Source or Object
#       form, that is based on (or derived from) the Work and for which the
#       editorial revisions, annotations, elaborations, or other modifications
#       represent, as a whole, an original work of authorship. For the purposes
#       of this License, Derivative Works shall not include works that remain
#       separable from, or merely link (or bind by name) to the interfaces of,
#       the Work and Derivative Works thereof.
#
#       "Contribution" shall mean any work of authorship, including
#       the original version of the Work and any modifications or additions
#       to that Work or Derivative Works thereof, that is intentionally
#       submitted to Licensor for inclusion in the Work by the copyright owner
#       or by an individual or Legal Entity authorized to submit on behalf of
#       the copyright owner. For the purposes of this definition, "submitted"
#       means any form of electronic, verbal, or written communication sent
#       to the Licensor or its representatives, including but not limited to
#       communication on electronic mailing lists, source code control systems,
#       and issue tracking systems that are managed by, or on behalf of, the
#       Licensor for the purpose of discussing and improving the Work, but
#       excluding communication that is conspicuously marked or otherwise
#       designated in writing by the copyright owner as "Not a Contribution."
#
#       "Contributor" shall mean Licensor and any individual or Legal Entity
#       on behalf of whom a Contribution has been received by Licensor and
#       subsequently incorporated within the Work.
#
#    2. Grant of Copyright License. Subject to the terms and conditions of
#       this License, each Contributor hereby grants to You a perpetual,
#       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#       copyright license to reproduce, prepare Derivative Works of,
#       publicly display, publicly perform, sublicense, and distribute the
#       Work and such Derivative Works in Source or Object form.
#
#    3. Grant of Patent License. Subject to the terms and conditions of
#       this License, each Contributor hereby grants to You a perpetual,
#       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#       (except as stated in this section) patent license to make, have made,
#       use, offer to sell, sell, import, and otherwise transfer the Work,
#       where such license applies only to those patent claims licensable
#       by such Contributor that are necessarily infringed by their
#       Contribution(s) alone or by combination of their Contribution(s)
#       with the Work to which such Contribution(s) was submitted. If You
#       institute patent litigation against any entity (including a
#       cross-claim or counterclaim in a lawsuit) alleging that the Work
#       or a Contribution incorporated within the Work constitutes direct
#       or contributory patent infringement, then any patent licenses
#       granted to You under this License for that Work shall terminate
#       as of the date such litigation is filed.
#
#    4. Redistribution. You may reproduce and distribute copies of the
#       Work or Derivative Works thereof in any medium, with or without
#       modifications, and in Source or Object form, provided that You
#       meet the following conditions:
#
#       (a) You must give any other recipients of the Work or
#           Derivative Works a copy of this License; and
#
#       (b) You must cause any modified files to carry prominent notices
#           stating that You changed the files; and
#
#       (c) You must retain, in the Source form of any Derivative Works
#           that You distribute, all copyright, patent, trademark, and
#           attribution notices from the Source form of the Work,
#           excluding those notices that do not pertain to any part of
#           the Derivative Works; and
#
#       (d) If the Work includes a "NOTICE" text file as part of its
#           distribution, then any Derivative Works that You distribute must
#           include a readable copy of the attribution notices contained
#           within such NOTICE file, excluding those notices that do not
#           pertain to any part of the Derivative Works, in at least one
#           of the following places: within a NOTICE text file distributed
#           as part of the Derivative Works; within the Source form or
#           documentation, if provided along with the Derivative Works; or,
#           within a display generated by the Derivative Works, if and
#           wherever such third-party notices normally appear. The contents
#           of the NOTICE file are for informational purposes only and
#           do not modify the License. You may add Your own attribution
#           notices within Derivative Works that You distribute, alongside
#           or as an addendum to the NOTICE text from the Work, provided
#           that such additional attribution notices cannot be construed
#           as modifying the License.
#
#       You may add Your own copyright statement to Your modifications and
#       may provide additional or different license terms and conditions
#       for use, reproduction, or distribution of Your modifications, or
#       for any such Derivative Works as a whole, provided Your use,
#       reproduction, and distribution of the Work otherwise complies with
#       the conditions stated in this License.
#
#    5. Submission of Contributions. Unless You explicitly state otherwise,
#       any Contribution intentionally submitted for inclusion in the Work
#       by You to the Licensor shall be under the terms and conditions of
#       this License, without any additional terms or conditions.
#       Notwithstanding the above, nothing herein shall supersede or modify
#       the terms of any separate license agreement you may have executed
#       with Licensor regarding such Contributions.
#
#    6. Trademarks. This License does not grant permission to use the trade
#       names, trademarks, service marks, or product names of the Licensor,
#       except as required for reasonable and customary use in describing the
#       origin of the Work and reproducing the content of the NOTICE file.
#
#    7. Disclaimer of Warranty. Unless required by applicable law or
#       agreed to in writing, Licensor provides the Work (and each
#       Contributor provides its Contributions) on an "AS IS" BASIS,
#       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#       implied, including, without limitation, any warranties or conditions
#       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
#       PARTICULAR PURPOSE. You are solely responsible for determining the
#       appropriateness of using or redistributing the Work and assume any
#       risks associated with Your exercise of permissions under this License.
#
#    8. Limitation of Liability. In no event and under no legal theory,
#       whether in tort (including negligence), contract, or otherwise,
#       unless required by applicable law (such as deliberate and grossly
#       negligent acts) or agreed to in writing, shall any Contributor be
#       liable to You for damages, including any direct, indirect, special,
#       incidental, or consequential damages of any character arising as a
#       result of this License or out of the use or inability to use the
#       Work (including but not limited to damages for loss of goodwill,
#       work stoppage, computer failure or malfunction, or any and all
#       other commercial damages or losses), even if such Contributor
#       has been advised of the possibility of such damages.
#
#    9. Accepting Warranty or Additional Liability. While redistributing
#       the Work or Derivative Works thereof, You may choose to offer,
#       and charge a fee for, acceptance of support, warranty, indemnity,
#       or other liability obligations and/or rights consistent with this
#       License. However, in accepting such obligations, You may act only
#       on Your own behalf and on Your sole responsibility, not on behalf
#       of any other Contributor, and only if You agree to indemnify,
#       defend, and hold each Contributor harmless for any liability
#       incurred by, or claims asserted against, such Contributor by reason
#       of your accepting any such warranty or additional liability.
#
#    END OF TERMS AND CONDITIONS
#
#    APPENDIX: How to apply the Apache License to your work.
#
#       To apply the Apache License to your work, attach the following
#       boilerplate notice, with the fields enclosed by brackets "{}"
#       replaced with your own identifying information. (Don't include
#       the brackets!)  The text should be enclosed in the appropriate
#       comment syntax for the file format. We also recommend that a
#       file or class name and description of purpose be included on the
#       same "printed page" as the copyright notice for easier
#       identification within third-party archives.
#
#    Copyright {yyyy} {name of copyright owner}
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# This code was modified by Francois Gallard, IRT Saint Exupery
# for integration within |g|

"""
Self Organizing Maps
********************
"""
from __future__ import  division, unicode_literals

import itertools
import logging
import os
import tempfile
from timeit import default_timer as timer

import matplotlib
import numpy as np
from matplotlib import cm
from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm
from scipy.sparse import csr_matrix
from sklearn import neighbors

try:
    from sklearn.externals.joblib import Parallel, delayed, dump, load
except ImportError:
    from joblib.parallel import Parallel, delayed

    dump = None
    load = None




try:
    from sklearn.decomposition import RandomizedPCA
except ImportError:
    from sklearn.decomposition import PCA as RandomizedPCA


LOGGER = logging.getLogger(__name__)


class SOM(object):
    """ """

    def __init__(
        self,
        name,
        Data,
        mapsize=None,
        norm_method="var",
        initmethod="pca",
        neigh="Guassian",
    ):
        """
        name and data, neigh== Bubble or Guassian
        """
        self.name = name
        self.data_raw = Data
        if norm_method == "var":
            Data = normalize(Data, method=norm_method)
            self.data = Data

        else:
            self.data = Data
        self.dim = Data.shape[1]
        self.dlen = Data.shape[0]
        self.set_topology(mapsize=mapsize)
        self.set_algorithm(initmethod=initmethod)
        self.calc_map_dist()
        self.neigh = neigh

        # Slow for large data sets
        # self.set_data_labels()

    # set SOM topology
    def set_topology(
        self, mapsize=None, mapshape="planar", lattice="rect", mask=None, compname=None
    ):
        """all_mapshapes = ['planar','toroid','cylinder']
        all_lattices = ['hexa','rect']

        :param mapsize: Default value = None)
        :param mapshape: Default value = 'planar')
        :param lattice: Default value = 'rect')
        :param mask: Default value = None)
        :param compname: Default value = None)

        """
        self.mapshape = mapshape
        self.lattice = lattice

        # to set mask
        if mask is None:
            self.mask = np.ones([1, self.dim])
        else:
            self.mask = mask

        # to set map size
        if mapsize is None:
            tmp = int(round(np.sqrt(self.dlen)))
            self.nnodes = tmp
            self.mapsize = [int(3.0 / 5 * self.nnodes), int(2.0 / 5 * self.nnodes)]
        else:
            if len(mapsize) == 2:
                if np.min(mapsize) == 1:
                    self.mapsize = [1, np.max(mapsize)]
                else:
                    self.mapsize = mapsize
            elif len(mapsize) == 1:
                s = int(mapsize[0] / 2)
                self.mapsize = [1, mapsize[0]]
                LOGGER.debug("input was considered as the numbers of nodes")
                LOGGER.debug("map size is [{0},{1}]".format(s, s))
            self.nnodes = self.mapsize[0] * self.mapsize[1]

        # to set discipline names
        if compname is None:
            try:
                cc = list()
                for i in range(0, self.dim):
                    cc.append("Variable-" + str(i + 1))
                    self.compname = np.asarray(cc)[np.newaxis, :]
            except Exception:
                pass
                LOGGER.debug("no data yet: plesae first set trainign data to the SOM")
        else:
            try:
                dim = getattr(self, "dim")
                if len(compname) == dim:
                    self.compname = np.asarray(compname)[np.newaxis, :]
                else:
                    LOGGER.debug("compname should have the same size")
            except Exception:
                pass
                LOGGER.debug("no data yet: plesae first set trainign data to the SOM")

    # Set labels of the training data
    # it should be in the format of a list of strings
    def set_data_labels(self, dlabel=None):
        """

        :param dlabel: Default value = None)

        """
        if dlabel is None:
            try:
                dlen = getattr(self, "dlen")
                cc = list()
                for i in range(0, dlen):
                    cc.append("dlabel-" + str(i))
                    self.dlabel = np.asarray(cc)[:, np.newaxis]
            except Exception:
                pass
                LOGGER.debug("no data yet: plesae first set trainign data to the SOM")
        else:
            try:
                dlen = getattr(self, "dlen")
                if dlabel.shape == (1, dlen):
                    self.dlabel = dlabel.T  # [:,np.newaxis]
                elif dlabel.shape == (dlen, 1):
                    self.dlabel = dlabel
                elif dlabel.shape == (dlen,):
                    self.dlabel = dlabel[:, np.newaxis]
                else:
                    LOGGER.debug("wrong lable format")
            except Exception:
                pass
                LOGGER.debug("no data yet: plesae first set trainign data to the SOM")

    # calculating the grid distance, which will be called during the training steps
    # currently just works for planar grids
    def calc_map_dist(self):
        """ """
        cd = getattr(self, "nnodes")
        UD2 = np.zeros((cd, cd))
        for i in range(cd):
            UD2[i, :] = grid_dist(self, i).reshape(1, cd)
        self.UD2 = UD2

    def set_algorithm(
        self,
        initmethod="pca",
        algtype="batch",
        neighborhoodmethod="gaussian",
        alfatype="inv",
        alfaini=0.5,
        alfafinal=0.005,
    ):
        """initmethod = ['random', 'pca']
        algos = ['seq','batch']
        all_neigh = ['gaussian','manhatan','bubble','cut_gaussian','epanechicov' ]
        alfa_types = ['linear','inv','power']

        :param initmethod: Default value = 'pca')
        :param algtype: Default value = 'batch')
        :param neighborhoodmethod: Default value = 'gaussian')
        :param alfatype: Default value = 'inv')
        :param alfaini: Default value = .5)
        :param alfafinal: Default value = .005)

        """
        self.initmethod = initmethod
        self.algtype = algtype
        self.alfaini = alfaini
        self.alfafinal = alfafinal
        self.neigh = neighborhoodmethod

    ###################################
    # visualize map
    def view_map(
        self,
        what="codebook",
        which_dim="all",
        pack="Yes",
        text_size=2.8,
        save="No",
        save_dir="empty",
        grid="No",
        text="Yes",
        cmap="None",
        COL_SiZe=6,
    ):
        """

        :param what: Default value = 'codebook')
        :param which_dim: Default value = 'all')
        :param pack: Default value = 'Yes')
        :param text_size: Default value = 2.8)
        :param save: Default value = 'No')
        :param save_dir: Default value = 'empty')
        :param grid: Default value = 'No')
        :param text: Default value = 'Yes')
        :param cmap: Default value = 'None')
        :param COL_SiZe: Default value = 6)

        """

        mapsize = getattr(self, "mapsize")
        if np.min(mapsize) > 1:
            if pack == "No":
                view_2d(self, text_size, which_dim=which_dim, what=what)
            else:
                #         		LOGGER.debug('hi')
                view_2d_Pack(
                    self,
                    text_size,
                    which_dim=which_dim,
                    what=what,
                    save=save,
                    save_dir=save_dir,
                    grid=grid,
                    text=text,
                    CMAP=cmap,
                    col_sz=COL_SiZe,
                )

        elif np.min(mapsize) == 1:
            view_1d(self, text_size, which_dim=which_dim, what=what)

    ##########################################################################
    # Initialize map codebook: Weight vectors of SOM
    def init_map(self):
        """ """
        if getattr(self, "initmethod") == "random":
            # It produces random values in the range of min- max of each
            # dimension based on a uniform distribution
            mn = np.tile(
                np.min(getattr(self, "data"), axis=0), (getattr(self, "nnodes"), 1)
            )
            mx = np.tile(
                np.max(getattr(self, "data"), axis=0), (getattr(self, "nnodes"), 1)
            )
            setattr(
                self,
                "codebook",
                mn
                + (mx - mn)
                * (np.random.rand(getattr(self, "nnodes"), getattr(self, "dim"))),
            )
        elif getattr(self, "initmethod") == "pca":
            # it is based on two largest eigenvalues of correlation matrix
            codebooktmp = lininit(self)
            setattr(self, "codebook", codebooktmp)
        else:
            LOGGER.debug("please select a corect initialization method")
            LOGGER.debug(
                "set a correct one in SOM. current SOM.initmethod:  ",
                getattr(self, "initmethod"),
            )
            LOGGER.debug("possible init methods:'random', 'pca'")

    # Main loop of training
    def train(self, trainlen=None, n_job=1, shared_memory="no", verbose="on"):
        """

        :param trainlen: Default value = None)
        :param n_job: Default value = 1)
        :param shared_memory: Default value = 'no')
        :param verbose: Default value = 'on')

        """
        # LOGGER.debug('data len is %d and data dimension is %d' % (dlen+str( dim)))
        # LOGGER.debug( 'map size is %d, %d' %(mapsize[0], mapsize[1]))
        # LOGGER.debug('array size in log10 scale' +str( mem))
        # LOGGER.debug('nomber of jobs in parallel: '+str( n_job))
        #######################################
        # initialization
        LOGGER.debug(
            "initialization method = %s, initializing.." % getattr(self, "initmethod")
        )
        t0 = timer()
        self.init_map()
        LOGGER.debug("initialization done in %f seconds" % round(timer() - t0))

        batchtrain(self, njob=n_job, phase="rough", shared_memory="no", verbose=verbose)
        batchtrain(
            self, njob=n_job, phase="finetune", shared_memory="no", verbose=verbose
        )
        err = np.mean(getattr(self, "bmu")[1])
        ts = round(timer() - t0, 3)
        LOGGER.debug("Total time elapsed: %f secodns" % ts)
        LOGGER.debug("final quantization error: %f" % err)
        ts = round(timer() - t0, 3)
        LOGGER.debug("Total time elapsed: %f secodns" % ts)
        LOGGER.debug("final quantization error: %f" % err)

    # to project a data set to a trained SOM and find the index of bmu
    # It is based on nearest neighborhood search module of scikitlearn, but it
    # is not that fast.
    def project_data(self, data):
        """

        :param data:

        """
        codebook = getattr(self, "codebook")
        data_raw = getattr(self, "data_raw")
        clf = neighbors.KNeighborsClassifier(n_neighbors=1)
        labels = np.arange(0, codebook.shape[0])
        clf.fit(codebook, labels)

        # the codebook values are all normalized
        # we can normalize the input data based on mean and std of original
        # data
        data = normalize_by(data_raw, data, method="var")
        # data = normalize(data, method='var')
        # plt.hist(data[:,2])
        Predicted_labels = clf.predict(data)
        return Predicted_labels

    def predict_by(self, data, Target, K=5, wt="distance"):
        """‘uniform’

        :param data: param Target:
        :param K: Default value = 5)
        :param wt: Default value = 'distance')
        :param Target:

        """
        # here it is assumed that Target is the last column in the codebook
        # and data has dim-1 columns
        codebook = getattr(self, "codebook")
        data_raw = getattr(self, "data_raw")
        dim = codebook.shape[1]
        ind = np.arange(0, dim)
        indX = ind[ind != Target]
        X = codebook[:, indX]
        Y = codebook[:, Target]
        n_neighbors = K
        clf = neighbors.KNeighborsRegressor(n_neighbors, weights=wt)
        clf.fit(X, Y)
        # the codebook values are all normalized
        # we can normalize the input data based on mean and std of original
        # data
        dimdata = data.shape[1]
        if dimdata == dim:
            data[:, Target] = 0
            data = normalize_by(data_raw, data, method="var")
            data = data[:, indX]
        elif dimdata == dim - 1:
            data = normalize_by(data_raw[:, indX], data, method="var")
            # data = normalize(data, method='var')
        Predicted_values = clf.predict(data)
        Predicted_values = denormalize_by(data_raw[:, Target], Predicted_values)
        return Predicted_values

    def predict(self, X_test, K=5, wt="distance"):
        """‘uniform’

        :param X_test: param K:  (Default value = 5)
        :param wt: Default value = 'distance')
        :param K:  (Default value = 5)

        """
        # Similar to SKlearn we assume that we have X_tr, Y_tr and X_test
        # here it is assumed that Target is the last column in the codebook
        # and data has dim-1 columns
        codebook = getattr(self, "codebook")
        data_raw = getattr(self, "data_raw")
        Target = data_raw.shape[1] - 1
        X_train = codebook[:, :Target]
        Y_train = codebook[:, Target]
        n_neighbors = K
        clf = neighbors.KNeighborsRegressor(n_neighbors, weights=wt)
        clf.fit(X_train, Y_train)
        # the codebook values are all normalized
        # we can normalize the input data based on mean and std of original
        # data
        X_test = normalize_by(data_raw[:, :Target], X_test, method="var")
        Predicted_values = clf.predict(X_test)
        Predicted_values = denormalize_by(data_raw[:, Target], Predicted_values)
        return Predicted_values

    def find_K_nodes(self, data, K=5):
        """

        :param data: param K:  (Default value = 5)
        :param K:  (Default value = 5)

        """
        from sklearn.neighbors import NearestNeighbors

        # we find the k most similar nodes to the input vector
        codebook = getattr(self, "codebook")
        neigh = NearestNeighbors(n_neighbors=K)
        neigh.fit(codebook)
        data_raw = getattr(self, "data_raw")
        # the codebook values are all normalized
        # we can normalize the input data based on mean and std of original
        # data
        data = normalize_by(data_raw, data, method="var")
        return neigh.kneighbors(data)

    def ind_to_xy(self, bm_ind):
        """

        :param bm_ind:

        """
        msize = getattr(self, "mapsize")
        rows = msize[0]
        cols = msize[1]
        # bmu should be an integer between 0 to no_nodes
        out = np.zeros((bm_ind.shape[0], 3))
        out[:, 2] = bm_ind
        out[:, 0] = rows - 1 - bm_ind / cols
        out[:, 0] = bm_ind / cols
        out[:, 1] = bm_ind % cols
        return out.astype(int)

    def cluster(self, method="Kmeans", n_clusters=8):
        """

        :param method: Default value = 'Kmeans')
        :param n_clusters: Default value = 8)

        """
        import sklearn.cluster as clust

        km = clust.KMeans(n_clusters=n_clusters)
        labels = km.fit_predict(
            denormalize_by(self.data_raw, self.codebook, n_method="var")
        )
        setattr(self, "cluster_labels", labels)
        return labels

    def hit_map(self, data=None):
        """

        :param data: Default value = None)

        """
        # First Step: show the hitmap of all the training data

        #     	LOGGER.debug('None')
        data_tr = getattr(self, "data_raw")
        proj = self.project_data(data_tr)
        msz = getattr(self, "mapsize")
        coord = self.ind_to_xy(proj)

        # this is not an appropriate way, but it works
        coord[:, 1] = msz[0] - coord[:, 1]
        ###############################
        fig = plt.figure(figsize=(msz[1] / 5, msz[0] / 5))
        ax = fig.add_subplot(111)
        ax.xaxis.set_ticks([i for i in range(0, msz[1])])
        ax.yaxis.set_ticks([i for i in range(0, msz[0])])
        ax.xaxis.set_ticklabels([])
        ax.yaxis.set_ticklabels([])
        ax.grid(True, linestyle="-", linewidth=0.5)
        a = plt.hist2d(
            coord[:, 1],
            coord[:, 0],
            bins=(msz[1], msz[0]),
            alpha=0.0,
            norm=LogNorm(),
            cmap=cm.get_cmap("jet"),
        )
        # clbar  = plt.colorbar()
        x = np.arange(0.5, msz[1] + 0.5, 1)
        y = np.arange(0.5, msz[0] + 0.5, 1)
        X, Y = np.meshgrid(x, y)
        area = a[0].T * 12
        plt.scatter(
            X,
            Y,
            s=area,
            alpha=0.2,
            c="b",
            marker="o",
            cmap="jet",
            linewidths=3,
            edgecolor="r",
        )
        plt.scatter(
            X,
            Y,
            s=area,
            alpha=0.9,
            c="None",
            marker="o",
            cmap="jet",
            linewidths=3,
            edgecolor="r",
        )
        plt.xlim(0, msz[1])
        plt.ylim(0, msz[0])

        if data is not None:
            proj = self.project_data(data)
            msz = getattr(self, "mapsize")
            coord = self.ind_to_xy(proj)
            a = plt.hist2d(
                coord[:, 1],
                coord[:, 0],
                bins=(msz[1], msz[0]),
                alpha=0.0,
                norm=LogNorm(),
                cmap=cm.get_cmap("jet"),
            )
            # clbar  = plt.colorbar()
            x = np.arange(0.5, msz[1] + 0.5, 1)
            y = np.arange(0.5, msz[0] + 0.5, 1)
            X, Y = np.meshgrid(x, y)
            area = a[0].T * 50
            plt.scatter(
                X,
                Y,
                s=area,
                alpha=0.2,
                c="b",
                marker="o",
                cmap="jet",
                linewidths=3,
                edgecolor="r",
            )
            plt.scatter(
                X,
                Y,
                s=area,
                alpha=0.9,
                c="None",
                marker="o",
                cmap="jet",
                linewidths=3,
                edgecolor="r",
            )
            plt.xlim(0, msz[1])
            plt.ylim(0, msz[0])

        plt.show()

    def hit_map_cluster_number(self, data=None):
        """

        :param data: Default value = None)

        """
        if hasattr(self, "cluster_labels"):
            codebook = getattr(self, "cluster_labels")
        #     		LOGGER.debug('yesyy')
        else:
            LOGGER.debug("clustering based on default parameters...")
            codebook = self.cluster()
        msz = getattr(self, "mapsize")
        fig = plt.figure(figsize=(msz[1] / 2.5, msz[0] / 2.5))
        ax = fig.add_subplot(111)
        #     	ax.xaxis.set_ticklabels([])
        #     	ax.yaxis.set_ticklabels([])
        #     	ax.grid(True,linestyle='-', linewidth=.5)

        if data is None:
            data_tr = getattr(self, "data_raw")
            proj = self.project_data(data_tr)
            cents = self.ind_to_xy(np.arange(0, msz[0] * msz[1]))
            for i, txt in enumerate(codebook):
                ax.annotate(txt, (cents[i, 1], cents[i, 0]), size=10, va="center")

        if data is not None:
            proj = self.project_data(data)
            cents = self.ind_to_xy(proj)
            #     		cents[:,1] = cents[:,1]+.2
            #     		LOGGER.debug(cents.shape)
            label = codebook[proj]
            for i, txt in enumerate(label):
                ax.annotate(txt, (cents[i, 1], cents[i, 0]), size=10, va="center")

        plt.imshow(codebook.reshape(msz[0], msz[1])[::], alpha=0.5)
        #     	plt.pcolor(codebook.reshape(msz[0],msz[1])[::-1],alpha=.5,cmap='jet')
        plt.show()
        return cents

    def predict_Probability(self, data, Target, K=5):
        """

        :param data: param Target:
        :param K: Default value = 5)
        :param Target:

        """
        # here it is assumed that Target is the last column in the codebook
        # #and data has dim-1 columns
        codebook = getattr(self, "codebook")
        data_raw = getattr(self, "data_raw")
        dim = codebook.shape[1]
        ind = np.arange(0, dim)
        indX = ind[ind != Target]
        X = codebook[:, indX]
        Y = codebook[:, Target]
        n_neighbors = K
        clf = neighbors.KNeighborsRegressor(n_neighbors, weights="distance")
        clf.fit(X, Y)
        # the codebook values are all normalized
        # we can normalize the input data based on mean and std of original
        # data
        dimdata = data.shape[1]
        if dimdata == dim:
            data[:, Target] = 0
            data = normalize_by(data_raw, data, method="var")
            data = data[:, indX]
        elif dimdata == dim - 1:
            data = normalize_by(data_raw[:, indX], data, method="var")
            # data = normalize(data, method='var')
        weights, ind = clf.kneighbors(data, n_neighbors=K, return_distance=True)
        weights = 1.0 / weights
        sum_ = np.sum(weights, axis=1)
        weights = weights / sum_[:, np.newaxis]
        labels = np.sign(codebook[ind, Target])
        labels[labels >= 0] = 1

        # for positives
        pos_prob = labels.copy()
        pos_prob[pos_prob < 0] = 0
        pos_prob = pos_prob * weights
        pos_prob = np.sum(pos_prob, axis=1)[:, np.newaxis]

        # for negatives
        neg_prob = labels.copy()
        neg_prob[neg_prob > 0] = 0
        neg_prob = neg_prob * weights * -1
        neg_prob = np.sum(neg_prob, axis=1)[:, np.newaxis]

        # Predicted_values = clf.predict(data)
        # Predicted_values = denormalize_by(data_raw[:,Target], Predicted_values)
        return np.concatenate((pos_prob, neg_prob), axis=1)

    def node_Activation(self, data, wt="distance", Target=None):
        """‘uniform’

        :param data: param wt:  (Default value = 'distance')
        :param Target: Default value = None)
        :param wt:  (Default value = 'distance')

        """

        if Target is None:
            codebook = getattr(self, "codebook")
            data_raw = getattr(self, "data_raw")
            clf = neighbors.KNeighborsClassifier(n_neighbors=getattr(self, "nnodes"))
            labels = np.arange(0, codebook.shape[0])
            clf.fit(codebook, labels)
            # the codebook values are all normalized
            # we can normalize the input data based on mean and std of original
            # data
            data = normalize_by(data_raw, data, method="var")
            weights, ind = clf.kneighbors(data)

            # Softmax function
            weights = 1.0 / weights
            S_ = np.sum(np.exp(weights), axis=1)[:, np.newaxis]
            weights = np.exp(weights) / S_

        return weights, ind

        #

    def para_bmu_find(self, x, y, njb=1):
        """

        :param x: param y:
        :param njb: Default value = 1)
        :param y:

        """
        dlen = x.shape[0]
        y_2 = np.einsum("ij,ij->i", y, y)
        # here it finds BMUs for chunk of data in parallel
        b = Parallel(n_jobs=njb, pre_dispatch="3*n_jobs")(
            delayed(chunk_based_bmu_find)(
                self, x[i * dlen // njb : min((i + 1) * dlen // njb, dlen)], y, y_2
            )
            for i in range(njb)
        )

        # LOGGER.debug('bmu finding: %f seconds ' %round(timer() - t_temp+str( 3)))
        bmu = np.asarray(list(itertools.chain(*b))).T
        # LOGGER.debug('bmu to array: %f seconds' %round(timer() - t1+str( 3)))
        del b
        return bmu

    # First finds the Voronoi set of each node. It needs to calculate a smaller matrix. Super fast comparing to classic batch training algorithm
    # it is based on the implemented algorithm in som toolbox for Matlab by
    # Helsinky university
    def update_codebook_voronoi(self, training_data, bmu, H, radius):
        """

        :param training_data: param bmu:
        :param H: param radius:
        :param bmu:
        :param radius:

        """
        # bmu has shape of 2,dlen, where first row has bmuinds
        # we construct ud2 from precomputed UD2: ud2 = UD2[bmu[0,:]]
        nnodes = getattr(self, "nnodes")
        dlen = getattr(self, "dlen")
        inds = bmu[0].astype(int)
        #         LOGGER.debug('bmu'+str( bmu[0]))
        #         fig = plt.hist(bmu[0],bins=100)
        #         plt.show()
        row = inds
        col = np.arange(dlen)
        val = np.tile(1, dlen)
        P = csr_matrix((val, (row, col)), shape=(nnodes, dlen))
        S = P.dot(training_data)
        # assert( S.shape == (nnodes, dim))
        # assert( H.shape == (nnodes, nnodes))

        # H has nnodes*nnodes and S has nnodes*dim  ---> Nominator has nnodes*dim
        # LOGGER.debug(Nom)
        Nom = H.T.dot(S)
        # assert( Nom.shape == (nnodes, dim))
        nV = P.sum(axis=1).reshape(1, nnodes)
        #         LOGGER.debug('nV'+str( nV))
        #         LOGGER.debug('H')
        #         LOGGER.debug( H)
        # assert(nV.shape == (1, nnodes))
        Denom = nV.dot(H.T).reshape(nnodes, 1)
        #         LOGGER.debug('Denom')
        #         LOGGER.debug( Denom)
        # assert( Denom.shape == (nnodes, 1))
        New_Codebook = np.divide(Nom, Denom)
        #         LOGGER.debug('codebook')
        #         LOGGER.debug(New_Codebook.sum(axis=1))
        Nom = None
        Denom = None
        # assert (New_Codebook.shape == (nnodes,dim))
        # setattr(som, 'codebook', New_Codebook)
        return np.around(New_Codebook, decimals=6)


# we will call this function in parallel for different number of jobs


def chunk_based_bmu_find(self, x, y, y_2):
    """

    :param x: param y:
    :param y_2:
    :param y:

    """
    dlen = x.shape[0]
    nnodes = y.shape[0]
    bmu = np.empty((dlen, 2))
    # it seems that smal batches for large dlen is really faster:
    # that is because of ddata in loops and n_jobs. for large data it slows
    # down due to memory needs in parallel
    blen = min(50, dlen)
    i0 = 0
    d = None
    while i0 + 1 <= dlen:
        Low = i0
        High = min(dlen, i0 + blen)
        i0 = i0 + blen
        ddata = x[Low : High + 1]
        d = np.dot(y, ddata.T)
        d *= -2
        d += y_2.reshape(nnodes, 1)
        bmu[Low : High + 1, 0] = np.argmin(d, axis=0)
        bmu[Low : High + 1, 1] = np.min(d, axis=0)
        del ddata
        d = None
    return bmu


# Batch training which is called for rought training as well as finetuning


def batchtrain(self, njob=1, phase=None, shared_memory="no", verbose="on"):
    """

    :param njob: Default value = 1)
    :param phase: Default value = None)
    :param shared_memory: Default value = 'no')
    :param verbose: Default value = 'on')

    """
    nnodes = getattr(self, "nnodes")
    dlen = getattr(self, "dlen")
    mapsize = getattr(self, "mapsize")

    #############################################
    # seting the parameters
    initmethod = getattr(self, "initmethod")
    mn = np.min(mapsize)
    if mn == 1:
        mpd = float(nnodes * 10) / float(dlen)
    else:
        mpd = float(nnodes) / float(dlen)

    ms = max(mapsize[0], mapsize[1])
    if mn == 1:
        ms = ms / 5.0
    # Based on somtoolbox, Matlab
    # case 'train',    sTrain.trainlen = ceil(50*mpd);
    # case 'rough',    sTrain.trainlen = ceil(10*mpd);
    # case 'finetune', sTrain.trainlen = ceil(40*mpd);
    if phase == "rough":
        # training length
        trainlen = int(np.ceil(30 * mpd))
        # radius for updating
        if initmethod == "random":
            radiusin = max(1, np.ceil(ms / 3.0))
            radiusfin = max(1, radiusin / 6.0)
        #         	radiusin = max(1, np.ceil(ms/1.))
        #         	radiusfin = max(1, radiusin/2.)
        elif initmethod == "pca":
            radiusin = max(1, np.ceil(ms / 8.0))
            radiusfin = max(1, radiusin / 4.0)
    elif phase == "finetune":
        # train lening length
        trainlen = int(np.ceil(60 * mpd))
        # radius for updating
        if initmethod == "random":
            trainlen = int(np.ceil(50 * mpd))
            radiusin = max(1, ms / 12.0)  # from radius fin in rough training
            radiusfin = max(1, radiusin / 25.0)

        # radiusin = max(1, ms/2.) #from radius fin in rough training
        #             radiusfin = max(1, radiusin/2.)
        elif initmethod == "pca":
            radiusin = max(1, np.ceil(ms / 8.0) / 4)
            radiusfin = 1  # max(1, ms/128)

    radius = np.linspace(radiusin, radiusfin, trainlen)
    ##################################################

    UD2 = getattr(self, "UD2")
    New_Codebook_V = getattr(self, "codebook")

    # LOGGER.debug('data is in shared memory?'+str( shared_memory))
    if shared_memory == "yes":
        data = getattr(self, "data")
        Data_folder = tempfile.mkdtemp()
        data_name = os.path.join(Data_folder, "data")
        dump(data, data_name)
        data = load(data_name, mmap_mode="r")
    else:
        data = getattr(self, "data")
    # x_2 is part of euclidean distance (x-y)^2 = x^2 +y^2 - 2xy that we use for each data row in bmu finding.
    # Since it is a fixed value we can skip it during bmu finding for each
    # data point, but later we need it calculate quantification error
    x_2 = np.einsum("ij,ij->i", data, data)
    if verbose == "on":
        LOGGER.debug("%s training..." % phase)
        LOGGER.debug(
            "radius_ini: %f , radius_final: %f, trainlen: %d"
            % (radiusin, radiusfin, trainlen)
        )
    neigh_func = getattr(self, "neigh")
    for i in range(trainlen):
        if neigh_func == "Guassian":
            # in case of Guassian neighborhood
            H = np.exp(-1.0 * UD2 / (2.0 * radius[i] ** 2)).reshape(nnodes, nnodes)
        if neigh_func == "Bubble":
            # in case of Bubble function
            #         	LOGGER.debug(radius[i]+str( UD2.shape))
            #         	LOGGER.debug(UD2)
            H = (
                l(radius[i], np.sqrt(UD2.flatten())).reshape(nnodes, nnodes)
                + 0.000000000001
            )
        #         	LOGGER.debug(H)
        t1 = timer()
        bmu = None
        bmu = self.para_bmu_find(data, New_Codebook_V, njb=njob)
        New_Codebook_V = self.update_codebook_voronoi(data, bmu, H, radius)
        # print 'updating nodes: ', round (timer()- t2, 3)
        if verbose == "on":
            LOGGER.debug(
                "epoch: %d ---> elapsed time:  %f, quantization error: %f "
                % (i + 1, round(timer() - t1, 3), np.mean(np.sqrt(bmu[1] + x_2)))
            )
    setattr(self, "codebook", New_Codebook_V)
    bmu[1] = np.sqrt(bmu[1] + x_2)
    setattr(self, "bmu", bmu)


def grid_dist(self, bmu_ind):
    """som and bmu_ind
    depending on the lattice "hexa" or "rect" we have different grid distance
    functions.
    bmu_ind is a number between 0 and number of nodes-1. depending on the map size
    bmu_coord will be calculated and then distance matrix in the map will be returned

    :param bmu_ind:

    """
    try:
        lattice = getattr(self, "lattice")
    except Exception:
        lattice = "hexa"
        LOGGER.debug("lattice not found! Lattice as hexa was set")

    if lattice == "rect":
        return rect_dist(self, bmu_ind)
    elif lattice == "hexa":
        try:
            msize = getattr(self, "mapsize")
            rows = msize[0]
            cols = msize[1]
        except Exception:
            rows = 0.0
            cols = 0.0
            pass

        # needs to be implemented
        LOGGER.debug("to be implemented" + str(rows) + str(cols))
        return np.zeros((rows, cols))


def rect_dist(self, bmu):
    """

    :param bmu:

    """
    # the way we consider the list of nodes in a planar grid is that node0 is on top left corner,
    # nodemapsz[1]-1 is top right corner and then it goes to the second row.
    # no. of rows is map_size[0] and no. of cols is map_size[1]
    try:
        msize = getattr(self, "mapsize")
        rows = msize[0]
        cols = msize[1]
    except Exception:
        pass

    # bmu should be an integer between 0 to no_nodes
    if 0 <= bmu <= (rows * cols):
        c_bmu = int(bmu % cols)
        r_bmu = int(bmu / cols)
    else:
        LOGGER.debug("wrong bmu")

    # calculating the grid distance
    if np.logical_and(rows > 0, cols > 0):
        r, c = np.arange(0, rows, 1)[:, np.newaxis], np.arange(0, cols, 1)
        dist2 = (r - r_bmu) ** 2 + (c - c_bmu) ** 2
        return dist2.ravel()
    else:
        LOGGER.debug("please consider the above mentioned errors")
        return np.zeros((rows, cols)).ravel()


def view_2d(self, text_size, which_dim="all", what="codebook"):
    """

    :param text_size: param which_dim:  (Default value = 'all')
    :param what: Default value = 'codebook')
    :param which_dim:  (Default value = 'all')

    """
    msz0, msz1 = getattr(self, "mapsize")
    if what == "codebook":
        if hasattr(self, "codebook"):
            codebook = getattr(self, "codebook")
            data_raw = getattr(self, "data_raw")
            codebook = denormalize_by(data_raw, codebook)
        else:
            LOGGER.debug("first initialize codebook")
        if which_dim == "all":
            dim = getattr(self, "dim")
            indtoshow = np.arange(0, dim).T
            ratio = float(dim) / float(dim)
            ratio = np.max((0.35, ratio))
            sH, sV = 16, 16 * ratio * 1
            plt.figure(figsize=(sH, sV))
        elif isinstance(which_dim, int):
            dim = 1
            indtoshow = np.zeros(1)
            indtoshow[0] = int(which_dim)
            sH, sV = 6, 6
            plt.figure(figsize=(sH, sV))
        elif isinstance(which_dim, list):
            max_dim = codebook.shape[1]
            dim = len(which_dim)
            ratio = float(dim) / float(max_dim)
            # print max_dim, dim, ratio
            ratio = np.max((0.35, ratio))
            indtoshow = np.asarray(which_dim).T
            sH, sV = 16, 16 * ratio * 1
            plt.figure(figsize=(sH, sV))

        no_row_in_plot = dim / 6 + 1  # 6 is arbitrarily selected
        if no_row_in_plot <= 1:
            no_col_in_plot = dim
        else:
            no_col_in_plot = 6

        axisNum = 0
        compname = getattr(self, "compname")
        norm = matplotlib.colors.normalize(
            vmin=np.mean(codebook.flatten()) - 1 * np.std(codebook.flatten()),
            vmax=np.mean(codebook.flatten()) + 1 * np.std(codebook.flatten()),
            clip=True,
        )
        while axisNum < dim:
            axisNum += 1
            ax = plt.subplot(no_row_in_plot, no_col_in_plot, axisNum)
            ind = int(indtoshow[axisNum - 1])
            mp = codebook[:, ind].reshape(msz0, msz1)
            pl = plt.pcolor(mp[::-1], norm=norm)
            #             pl = plt.imshow(mp[::-1])
            plt.title(compname[0][ind])
            font = {"size": text_size * sH / no_col_in_plot}
            plt.rc("font", **font)
            plt.axis("off")
            plt.axis([0, msz0, 0, msz1])
            ax.set_yticklabels([])
            ax.set_xticklabels([])
            plt.colorbar(pl)
        plt.show()


def view_2d_Pack(
    self,
    text_size,
    which_dim="all",
    what="codebook",
    save="No",
    grid="Yes",
    save_dir="empty",
    text="Yes",
    CMAP="None",
    col_sz=None,
):
    """

    :param text_size: param which_dim:  (Default value = 'all')
    :param what: Default value = 'codebook')
    :param save: Default value = 'No')
    :param grid: Default value = 'Yes')
    :param save_dir: Default value = 'empty')
    :param text: Default value = 'Yes')
    :param CMAP: Default value = 'None')
    :param col_sz: Default value = None)
    :param which_dim:  (Default value = 'all')

    """
    msz0, msz1 = getattr(self, "mapsize")
    if CMAP == "None":
        CMAP = cm.get_cmap("jet")
    if what == "codebook":
        if hasattr(self, "codebook"):
            codebook = getattr(self, "codebook")
            data_raw = getattr(self, "data_raw")
            codebook = denormalize_by(data_raw, codebook)
        else:
            LOGGER.debug("first initialize codebook")
        if which_dim == "all":
            dim = getattr(self, "dim")
            indtoshow = np.arange(0, dim).T
            ratio = float(dim) / float(dim)
            ratio = np.max((0.35, ratio))
        #             sH, sV = 16, 16 * ratio * 1
        #             plt.figure(figsize=(sH,sV))
        elif isinstance(which_dim, int):
            dim = 1
            indtoshow = np.zeros(1)
            indtoshow[0] = int(which_dim)
        #             sH, sV = 6, 6
        #             plt.figure(figsize=(sH,sV))
        elif isinstance(which_dim, list):
            max_dim = codebook.shape[1]
            dim = len(which_dim)
            ratio = float(dim) / float(max_dim)
            # print max_dim, dim, ratio
            ratio = np.max((0.35, ratio))
            indtoshow = np.asarray(which_dim).T
        #             sH, sV = 16, 16 * ratio * 1
        #             plt.figure(figsize=(sH,sV))

        #         plt.figure(figsize=(7,7))
        no_row_in_plot = dim / col_sz + 1  # 6 is arbitrarily selected
        if no_row_in_plot <= 1:
            no_col_in_plot = dim
        else:
            no_col_in_plot = col_sz

        axisNum = 0
        compname = getattr(self, "compname")
        h = 0.1
        w = 0.1
        fig = plt.figure(
            figsize=(no_col_in_plot * 2.5 * (1 + w), no_row_in_plot * 2.5 * (1 + h))
        )
        #         LOGGER.debug(no_row_in_plot+str( no_col_in_plot))
        #         norm = matplotlib.colors.Normalize(vmin=np.median(codebook.flatten()) - 1.5 * np.std(
        #             codebook.flatten()), vmax=np.median(codebook.flatten()) + 1.5 * np.std(codebook.flatten()), clip=False)
        #
        #         DD = pd.Series(data=codebook.flatten()).describe(
        #             percentiles=[.03, .05, .1, .25, .3, .4, .5, .6, .7, .8, .9, .95, .97])
        # #         norm = matplotlib.colors.Normalize(
        #             vmin=DD.ix['3%'], vmax=DD.ix['97%'], clip=False)

        while axisNum < dim:
            axisNum += 1

            ax = fig.add_subplot(no_row_in_plot, no_col_in_plot, axisNum)
            ind = int(indtoshow[axisNum - 1])
            mp = codebook[:, ind].reshape(msz0, msz1)

            if grid == "Yes":
                pl = plt.pcolor(mp[::-1])
            elif grid == "No":
                plt.imshow(mp[::-1], norm=None, cmap=CMAP)
                #             	plt.pcolor(mp[::-1])
                plt.axis("off")

            if text == "Yes":
                plt.title(compname[0][ind])
                font = {"size": text_size}
                plt.rc("font", **font)
            plt.axis([0, msz0, 0, msz1])
            ax.set_yticklabels([])
            ax.set_xticklabels([])
            ax.xaxis.set_ticks([i for i in range(0, msz1)])
            ax.yaxis.set_ticks([i for i in range(0, msz0)])
            ax.xaxis.set_ticklabels([])
            ax.yaxis.set_ticklabels([])
            ax.grid(True, linestyle="-", linewidth=0.5, color="k")
        #             plt.grid()
        #             plt.colorbar(pl)
        #         plt.tight_layout()
        plt.subplots_adjust(hspace=h, wspace=w)
    if what == "cluster":
        if hasattr(self, "cluster_labels"):
            codebook = getattr(self, "cluster_labels")
        else:
            LOGGER.debug("clustering based on default parameters...")
            codebook = self.cluster()
        h = 0.2
        w = 0.001
        fig = plt.figure(figsize=(msz0 / 2, msz1 / 2))

        ax = fig.add_subplot(1, 1, 1)
        mp = codebook[:].reshape(msz0, msz1)
        if grid == "Yes":
            pl = plt.pcolor(mp[::-1])
        elif grid == "No":
            plt.imshow(mp[::-1])
            #             plt.pcolor(mp[::-1])
            plt.axis("off")

        if text == "Yes":
            plt.title("clusters")
            font = {"size": text_size}
            plt.rc("font", **font)
        plt.axis([0, msz0, 0, msz1])
        ax.set_yticklabels([])
        ax.set_xticklabels([])
        ax.xaxis.set_ticks([i for i in range(0, msz1)])
        ax.yaxis.set_ticks([i for i in range(0, msz0)])
        ax.xaxis.set_ticklabels([])
        ax.yaxis.set_ticklabels([])
        ax.grid(True, linestyle="-", linewidth=0.5, color="k")
        plt.subplots_adjust(hspace=h, wspace=w)

    if save == "Yes":

        if save_dir != "empty":
            #         		LOGGER.debug(save_dir)
            fig.savefig(save_dir, bbox_inches="tight", transparent=False, dpi=400)
        else:
            #         		LOGGER.debug(save_dir)
            add = "/Users/itadmin/Desktop/SOM.png"
            fig.savefig(add, bbox_inches="tight", transparent=False, dpi=400)

        plt.close(fig)


def view_1d(self, text_size, which_dim="all", what="codebook"):
    """

    :param text_size: param which_dim:  (Default value = 'all')
    :param what: Default value = 'codebook')
    :param which_dim:  (Default value = 'all')

    """
    #     msz0, msz1 = getattr(self, 'mapsize')
    if what == "codebook":
        if hasattr(self, "codebook"):
            codebook = getattr(self, "codebook")
            data_raw = getattr(self, "data_raw")
            codebook = denormalize_by(data_raw, codebook)
        else:
            LOGGER.debug("first initialize codebook")
        if which_dim == "all":
            dim = getattr(self, "dim")
            indtoshow = np.arange(0, dim).T
            ratio = float(dim) / float(dim)
            ratio = np.max((0.35, ratio))
            sH, sV = 16, 16 * ratio * 1
            plt.figure(figsize=(sH, sV))
        elif isinstance(which_dim, int):
            dim = 1
            indtoshow = np.zeros(1)
            indtoshow[0] = int(which_dim)
            sH, sV = 6, 6
            plt.figure(figsize=(sH, sV))
        elif isinstance(which_dim, list):
            max_dim = codebook.shape[1]
            dim = len(which_dim)
            ratio = float(dim) / float(max_dim)
            # print max_dim, dim, ratio
            ratio = np.max((0.35, ratio))
            indtoshow = np.asarray(which_dim).T
            sH, sV = 16, 16 * ratio * 1
            plt.figure(figsize=(sH, sV))

        no_row_in_plot = dim / 6 + 1  # 6 is arbitrarily selected
        if no_row_in_plot <= 1:
            no_col_in_plot = dim
        else:
            no_col_in_plot = 6

        axisNum = 0
        compname = getattr(self, "compname")
        while axisNum < dim:
            axisNum += 1
            plt.subplot(no_row_in_plot, no_col_in_plot, axisNum)
            ind = int(indtoshow[axisNum - 1])
            mp = codebook[:, ind]
            plt.plot(mp, "-k", linewidth=0.8)
            # pl = plt.pcolor(mp[::-1])
            plt.title(compname[0][ind])
            font = {"size": text_size * sH / no_col_in_plot}
            plt.rc("font", **font)
            # plt.axis('off')
            # plt.axis([0, msz0, 0, msz1])
            # ax.set_yticklabels([])
            # ax.set_xticklabels([])
            # plt.colorbar(pl)
        plt.show()


def lininit(self):
    """ """
    # X = UsigmaWT
    # XTX = Wsigma^2WT
    # T = XW = Usigma #Transformed by W EigenVector, can be calculated by
    # multiplication PC matrix by eigenval too
    # Furthe, we can get lower ranks by using just few of the eigen vevtors
    # T(2) = U(2)sigma(2) = XW(2) ---> 2 is the number of selected eigenvectors
    # This is how we initialize the map, just by using the first two first eigen vals and eigenvectors
    # Further, we create a linear combination of them in the new map by giving values from -1 to 1 in each
    # Direction of SOM map
    # it shoud be noted that here, X is the covariance matrix of original data

    msize = getattr(self, "mapsize")
    #     rows = msize[0]
    cols = msize[1]
    nnodes = getattr(self, "nnodes")

    if np.min(msize) > 1:
        coord = np.zeros((nnodes, 2))
        for i in range(0, nnodes):
            coord[i, 0] = int(i / cols)  # x
            coord[i, 1] = int(i % cols)  # y
        mx = np.max(coord, axis=0)
        mn = np.min(coord, axis=0)
        coord = (coord - mn) / (mx - mn)
        coord = (coord - 0.5) * 2
        data = getattr(self, "data")
        me = np.mean(data, 0)
        data = data - me
        codebook = np.tile(me, (nnodes, 1))
        pca = RandomizedPCA(n_components=2)  # Randomized PCA is scalable
        # pca = PCA(n_components=2)
        pca.fit(data)
        eigvec = pca.components_
        eigval = pca.explained_variance_
        norms = np.sqrt(np.einsum("ij,ij->i", eigvec, eigvec))
        eigvec = ((eigvec.T / norms) * eigval).T
        eigvec.shape

        for j in range(nnodes):
            for i in range(eigvec.shape[0]):
                codebook[j, :] = codebook[j, :] + coord[j, i] * eigvec[i, :]
        return np.around(codebook, decimals=6)
    elif np.min(msize) == 1:
        coord = np.zeros((nnodes, 1))
        for i in range(0, nnodes):
            # coord[i,0] = int(i/cols) #x
            coord[i, 0] = int(i % cols)  # y
        mx = np.max(coord, axis=0)
        mn = np.min(coord, axis=0)
        # LOGGER.debug(coord)

        coord = (coord - mn) / (mx - mn)
        coord = (coord - 0.5) * 2
        # LOGGER.debug(coord)
        data = getattr(self, "data")
        me = np.mean(data, 0)
        data = data - me
        codebook = np.tile(me, (nnodes, 1))
        pca = RandomizedPCA(n_components=1)  # Randomized PCA is scalable
        # pca = PCA(n_components=2)
        pca.fit(data)
        eigvec = pca.components_
        eigval = pca.explained_variance_
        norms = np.sqrt(np.einsum("ij,ij->i", eigvec, eigvec))
        eigvec = ((eigvec.T / norms) * eigval).T
        eigvec.shape

        for j in range(nnodes):
            for i in range(eigvec.shape[0]):
                codebook[j, :] = codebook[j, :] + coord[j, i] * eigvec[i, :]
        return np.around(codebook, decimals=6)


def normalize(data, method="var"):
    """

    :param data: param method:  (Default value = 'var')
    :param method:  (Default value = 'var')

    """
    # methods  = ['var','range','log','logistic','histD','histC']
    # status = ['done', 'undone']
    me = np.mean(data, axis=0)
    st = np.std(data, axis=0)
    if method == "var":
        me = np.mean(data, axis=0)
        st = np.std(data, axis=0)
        n_data = (data - me) / st
        return n_data


def normalize_by(data_raw, data, method="var"):
    """

    :param data_raw: param data:
    :param method: Default value = 'var')
    :param data:

    """
    # methods  = ['var','range','log','logistic','histD','histC']
    # status = ['done', 'undone']
    # to have the mean and std of the original data, by which SOM is trained
    me = np.mean(data_raw, axis=0)
    st = np.std(data_raw, axis=0)
    if method == "var":
        n_data = (data - me) / st
        return n_data


def denormalize_by(data_by, n_vect, n_method="var"):
    """

    :param data_by: param n_vect:
    :param n_method: Default value = 'var')
    :param n_vect:

    """
    # based on the normalization
    if n_method == "var":
        me = np.mean(data_by, axis=0)
        st = np.std(data_by, axis=0)
        vect = n_vect * st + me
        return vect
    else:
        LOGGER.debug("data is not normalized before")
        return n_vect


def l(a, b):
    """

    :param a: param b:
    :param b:

    """
    c = np.zeros(b.shape)
    c[a - b >= 0] = 1
    return c


# Function to show hits
# som_labels = sm.project_data(Tr_Data)
# S = pd.DataFrame(data=som_labels,columns= ['label'])
# a = S['label'].value_counts()
# a = a.sort_index()
# a = pd.DataFrame(data=a.values, index=a.index,columns=['label'])
# d = pd.DataFrame(data= range(msz0*msz1),columns=['node_ID'])
# c  = d.join(a,how='outer')
# c.fillna(value=0,inplace=True)
# hits = c.values[:,1]
# hits = hits
# nodeID = np.arange(msz0*msz1)
# c_bmu = nodeID%msz1
# r_bmu = msz0 - nodeID/msz1
# fig, ax = plt.subplots()
# plt.axis([0, msz0, 0, msz1])
# ax.scatter(r_bmu, c_bmu, s=hits/2)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard, Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""The RBF network for regression.

The radial basis function surrogate discipline expresses the model output
as a weighted sum of kernel functions centered on the learning input data:

.. math::

    y = w_1K(\|x-x_1\|;\epsilon) + w_2K(\|x-x_2\|;\epsilon) + ...
        + w_nK(\|x-x_n\|;\epsilon)

and the coefficients :math:`(w_1, w_2, ..., w_n)` are estimated
by least squares minimization.

Dependence
----------
The RBF model relies on the Rbf class
of the `scipy library <https://docs.scipy.org/doc/scipy/reference/generated/
scipy.interpolate.Rbf.html>`_.
"""
from __future__ import division, unicode_literals

import logging
import pickle
from os.path import join
from typing import Callable, Dict, Iterable, Optional, Union

from numpy import array, average, exp, finfo, hstack, log, ndarray, sqrt
from numpy.linalg import norm
from scipy.interpolate import Rbf
from six import string_types

from gemseo.core.dataset import Dataset
from gemseo.mlearning.core.ml_algo import MLAlgoParameterType, TransformerType
from gemseo.mlearning.core.supervised import SavedObjectType
from gemseo.mlearning.regression.regression import MLRegressionAlgo
from gemseo.utils.py23_compat import PY3

LOGGER = logging.getLogger(__name__)

SavedObjectType = Union[SavedObjectType, float, Callable]


class RBFRegression(MLRegressionAlgo):
    r"""Regression based on radial basis functions.

    Attributes:
        der_function (Callable[[ndarray],ndarray]): The derivative
            of the radial basis function.
        y_average (ndarray): The mean of the learning output data.
    """

    LIBRARY = "scipy"
    ABBR = "RBF"

    EUCLIDEAN = "euclidean"

    MULTIQUADRIC = "multiquadric"
    INVERSE_MULTIQUADRIC = "inverse_multiquadric"
    GAUSSIAN = "gaussian"
    LINEAR = "linear"
    CUBIC = "cubic"
    QUINTIC = "quintic"
    THIN_PLATE = "thin_plate"

    AVAILABLE_FUNCTIONS = [
        MULTIQUADRIC,
        INVERSE_MULTIQUADRIC,
        GAUSSIAN,
        LINEAR,
        CUBIC,
        QUINTIC,
        THIN_PLATE,
    ]

    def __init__(
        self,
        data,  # type: Dataset
        transformer=None,  # type: Optional[TransformerType]
        input_names=None,  # type: Optional[Iterable[str]]
        output_names=None,  # type: Optional[Iterable[str]]
        function=MULTIQUADRIC,  # type: str
        der_function=None,  # type: Optional[Callable[[ndarray],ndarray]]
        epsilon=None,  # type: Optional[float]
        **parameters  # type: Optional[MLAlgoParameterType]
    ):  # type: (...) -> None
        r"""
        Args:
            function: The radial basis function.
            der_function: The derivative of the radial basis function,
                only to be provided if ``function`` is a callable
                and if the use of the model with its derivative is required.
                If None and if ``function`` is a callable,
                an error will be raised.
                If None and if ``function`` is a string,
                the class will look for its internal implementation
                and will raise an error if it is missing.
                The der_function shall take three arguments
                (input_data, norm_input_data, eps).
                For a RBF of the form function(:math:`r`),
                der_function(:math:`x`, :math:`|x|`, :math:`\epsilon`) shall
                return :math:`\epsilon^{-1} x/|x| f'(|x|/\epsilon)`.
            epsilon: An adjustable constant for Gaussian or multiquadrics functions.
                If None, use the average distance between input data.
        """
        if isinstance(function, string_types):
            function = str(function)
        super(RBFRegression, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            function=function,
            epsilon=epsilon,
            **parameters
        )
        self.y_average = 0.0
        self.der_function = der_function

    class RBFDerivatives(object):
        r"""Derivatives of functions used in :class:`.RBFRegression`.

        For a RBF of the form :math:`f(r)`, :math:`r` scalar,
        the derivative functions are defined by :math:`d(f(r))/dx`,
        with :math:`r=|x|/\epsilon`. The functions are thus defined
        by :math:`df/dx = \epsilon^{-1} x/|x| f'(|x|/\epsilon)`.
        This convention is chosen to avoid division by :math:`|x|` when
        the terms may be cancelled out, as :math:`f'(r)` often has a term
        in :math:`r`.
        """

        TOL = finfo(float).eps

        @classmethod
        def der_multiquadric(
            cls,
            input_data,  # type: ndarray
            norm_input_data,  # type: float
            eps,  # type: float
        ):  # type: (...) -> ndarray
            r"""Compute derivative of  :math:`f(r) = \sqrt{r^2 + 1}` wrt :math:`x`.

            Args:
                input_data: The 1D input data.
                norm_input_data: The norm of the input variable.
                eps: The correlation length.

            Returns:
                The derivative of the function.
            """
            return input_data / eps ** 2 / sqrt((norm_input_data / eps) ** 2 + 1)

        @classmethod
        def der_inverse_multiquadric(
            cls,
            input_data,  # type: ndarray
            norm_input_data,  # type: float
            eps,  # type: float
        ):  # type: (...) -> ndarray
            r"""Compute derivative w.r.t. :math:`x` of the function
            :math:`f(r) = 1/\sqrt{r^2 + 1}`.

            Args:
                input_data: The 1D input data.
                norm_input_data: The norm of the input variable.
                eps: The correlation length.

            Returns:
                The derivative of the function.
            """
            return -input_data / eps ** 2 / ((norm_input_data / eps) ** 2 + 1) ** 1.5

        @classmethod
        def der_gaussian(
            cls,
            input_data,  # type: ndarray
            norm_input_data,  # type: float
            eps,  # type: float
        ):  # type: (...) -> ndarray
            r"""Compute derivative w.r.t. :math:`x` of the function
            :math:`f(r) = \exp(-r^2)`.

            Args:
                input_data: The 1D input data.
                norm_input_data: The norm of the input variable.
                eps: The correlation length.

            Returns:
                The derivative of the function.
            """
            return -2 * input_data / eps ** 2 * exp(-((norm_input_data / eps) ** 2))

        @classmethod
        def der_linear(
            cls,
            input_data,  # type: ndarray
            norm_input_data,  # type: float
            eps,  # type: float
        ):  # type: (...) -> ndarray
            """Compute derivative w.r.t. :math:`x` of the function
            :math:`f(r) = r`.
            If :math:`x=0`, return 0 (determined up to a tolerance).

            Args:
                input_data: The 1D input data.
                norm_input_data: The norm of the input variable.
                eps: The correlation length.

            Returns:
                The derivative of the function.
            """
            return (
                (norm_input_data > cls.TOL)
                * input_data
                / eps
                / (norm_input_data + cls.TOL)
            )

        @classmethod
        def der_cubic(
            cls,
            input_data,  # type: ndarray
            norm_input_data,  # type: float
            eps,  # type: float
        ):  # type: (...) -> ndarray
            """Compute derivative w.r.t. :math:`x` of the function
            :math:`f(r) = r^3`.

            Args:
                input_data: The 1D input data.
                norm_input_data: The norm of the input variable.
                eps: The correlation length.

            Returns:
                The derivative of the function.
            """
            return 3 * norm_input_data * input_data / eps ** 3

        @classmethod
        def der_quintic(
            cls,
            input_data,  # type: ndarray
            norm_input_data,  # type: float
            eps,  # type: float
        ):  # type: (...) -> ndarray
            """Compute derivative w.r.t. :math:`x` of the function
            :math:`f(r) = r^5`.

            Args:
                input_data: The 1D input data.
                norm_input_data : The norm of the input variable.
                eps: The correlation length.

            Returns:
                The derivative of the function.
            """
            return 5 * norm_input_data ** 3 * input_data / eps ** 5

        @classmethod
        def der_thin_plate(
            cls,
            input_data,  # type: ndarray
            norm_input_data,  # type: float
            eps,  # type: float
        ):  # type: (...) -> ndarray
            r"""Compute derivative w.r.t. :math:`x` of the function
            :math:`f(r) = r^2 \log(r)`.
            If :math:`x=0`, return 0 (determined up to a tolerance).

            Args:
                input_data: The 1D input data.
                norm_input_data: The norm of the input variable.
                eps: The correlation length.

            Returns:
                The derivative of the function.
            """
            return (
                (norm_input_data > cls.TOL)
                * input_data
                / eps ** 2
                * (1 + 2 * log(norm_input_data / eps + cls.TOL))
            )

    def _fit(
        self,
        input_data,  # type: ndarray
        output_data,  # type: ndarray
    ):  # type: (...) -> None
        self.y_average = average(output_data, axis=0)
        output_data -= self.y_average
        if PY3:
            args = list(input_data.T) + [output_data]
            self.algo = Rbf(*args, mode="N-D", **self.parameters)
        else:
            self.algo = []
            for output in range(output_data.shape[1]):
                args = hstack([input_data, output_data[:, [output]]])
                rbf = Rbf(*args.T, **self.parameters)
                self.algo.append(rbf)

    def _predict(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        if PY3:
            output_data = self.algo(*input_data.T)
            if len(output_data.shape) == 1:
                output_data = output_data[:, None]  # n_outputs=1, rbf reduces
            output_data = output_data + self.y_average
        else:
            output_data = [rbf(*input_data.T) for rbf in self.algo]
            output_data = array(output_data).T + self.y_average
        return output_data

    def _predict_jacobian(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        self._check_available_jacobian()
        der_func = self.der_function or getattr(
            self.RBFDerivatives, "der_{}".format(self.function)
        )
        #             predict_samples                        learn_samples
        # Dimensions : ( n_samples , n_outputs , n_inputs , n_learn_samples )
        # input_data : ( n_samples ,           , n_inputs ,                 )
        # ref_points : (           ,           , n_inputs , n_learn_samples )
        # nodes      : (           , n_outputs ,          , n_learn_samples )
        # jacobians  : ( n_samples , n_outputs , n_inputs ,                 )
        if PY3:
            eps = self.algo.epsilon
            ref_points = self.algo.xi[None, None]
            nodes = self.algo.nodes.T[None, :, None]
        else:
            eps = [rbf.epsilon for rbf in self.algo]
            eps = array(eps)[None, :, None, None]  # 1 epsilon for each output
            ref_points = self.algo[0].xi  # same xi for all algos
            nodes = array([rbf.nodes for rbf in self.algo])[None, :, None]
        input_data = input_data[:, None, :, None]
        diffs = input_data - ref_points
        dists = norm(diffs, axis=2)[:, :, None]
        contributions = nodes * der_func(diffs, dists, eps=eps)
        jacobians = contributions.sum(-1)
        return jacobians

    def _check_available_jacobian(self):  # type: (...) -> None
        """Check if the Jacobian is available for the given setup.

        Raises:
            NotImplementedError: Either if the Jacobian computation is not implemented
                or if the derivative of the radial basis function is missing.
        """
        if PY3:
            norm_name = self.algo.norm
        else:
            norm_name = self.algo[0].norm

        if norm_name != self.EUCLIDEAN:
            raise NotImplementedError(
                "Jacobian is only implemented for Euclidean norm."
            )

        if callable(self.function) and self.der_function is None:
            raise NotImplementedError(
                "No der_function is provided."
                "Add der_function in RBFRegression constructor."
            )

    def _save_algo(
        self,
        directory,  # type: str
    ):  # type: (...) -> None
        if PY3:
            super(RBFRegression, self)._save_algo(directory)
        else:
            filename = join(directory, "algo.pkl")
            with open(filename, "wb") as handle:
                pickled_rbf = pickle.Pickler(handle, protocol=2)
                pickled_rbf_list = []
                for rbf in self.algo:
                    pickled_rbf_list.append({})
                    for key in rbf.__dict__.keys():
                        if key not in ["_function", "norm"]:
                            pickled_rbf_list[-1][key] = rbf.__getattribute__(key)
                pickled_rbf.dump(pickled_rbf_list)

    def load_algo(
        self,
        directory,  # type: str
    ):  # type: (...) -> None
        if PY3:
            super(RBFRegression, self).load_algo(directory)
        else:
            filename = join(directory, "algo.pkl")
            self.algo = []
            with open(filename, "rb") as handle:
                unpickled_rbf = pickle.Unpickler(handle)
                unpickled_rbf_list = unpickled_rbf.load()
                for rbf in unpickled_rbf_list:
                    algo_i = Rbf(
                        array([1, 2, 3]),
                        array([10, 20, 30]),
                        array([100, 200, 300]),
                        function=rbf["function"],
                    )
                    for key, value in rbf.items():
                        algo_i.__setattr__(key, value)
                    self.algo.append(algo_i)

    def _get_objects_to_save(self):  # type: (...) -> Dict[str,SavedObjectType]
        objects = super(RBFRegression, self)._get_objects_to_save()
        objects["y_average"] = self.y_average
        objects["der_function"] = self.der_function
        return objects

    @property
    def function(self):  # type: (...) -> str
        """The name of the kernel function.

        The name is possibly different from self.parameters['function'], as it
        is mapped (scipy). Examples:

        'inverse'              -> 'inverse_multiquadric'
        'InverSE MULtiQuadRIC' -> 'inverse_multiquadric'
        """
        if PY3:
            function = self.algo.function
        else:
            function = self.algo[0].function
        return function



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Computation of tolerance intervals from a data-fitted normal distribution."""

from typing import Tuple

import openturns as ot
from numpy import array, inf, ndarray
from past.utils import old_div

from gemseo.uncertainty.statistics.tolerance_interval.distribution import (
    ToleranceInterval,
    ToleranceIntervalSide,
)


class NormalToleranceInterval(ToleranceInterval):
    """Computation of tolerance intervals from a data-fitted normal distribution.

    The formulae come from the R library *tolerance* [1]_.

    .. [1] Derek S. Young,
           *tolerance: An R Package for Estimating Tolerance Intervals*,
           Journal of Statistical Software, 36(5), 2010
    """

    def __init__(
        self,
        size,  # type: int
        mean,  # type: float
        std,  # type: float
    ):  # type:(...) -> None
        # noqa: D205 D212 D415
        """
        Args:
            mean: The estimation of the mean of the normal distribution.
            std: The estimation of the standard deviation of the normal distribution.
        """
        super(NormalToleranceInterval, self).__init__(size)
        self.__mean = mean
        self.__std = std

    def _compute(
        self,
        coverage,  # type: float
        alpha,  # type: float
        size,  # type: int
        side,  # type: ToleranceIntervalSide
    ):  # type: (...) -> Tuple[ndarray,ndarray]
        if side in [
            ToleranceIntervalSide.UPPER,
            ToleranceIntervalSide.LOWER,
        ]:
            offset = ot.Normal().computeQuantile(coverage)[0] * size ** 0.5
            student = ot.Student(size - 1, offset)
            student_quantile = student.computeQuantile(1 - alpha)[0]
            tolerance_factor = old_div(student_quantile, size ** 0.5)

            if side == ToleranceIntervalSide.UPPER:
                upper = self.__mean + tolerance_factor * self.__std
                return array([-inf]), array([upper])
            else:
                lower = self.__mean - tolerance_factor * self.__std
                return array([lower]), array([inf])

        elif side == ToleranceIntervalSide.BOTH:
            z_p = ot.Normal().computeQuantile((1 + coverage) / 2.0)[0]
            u_term = (1 + 1.0 / size) ** 0.5 * z_p
            chi_square = ot.ChiSquare(size - 1)
            v_term = (old_div((size - 1), chi_square.computeQuantile(alpha)[0])) ** 0.5
            w_term = (
                1
                + (size - 3 - chi_square.computeQuantile(alpha)[0])
                / (2 * (size + 1) ** 2)
            ) ** 0.05
            tolerance_factor = u_term * v_term * w_term
            lower = self.__mean - tolerance_factor * self.__std
            upper = self.__mean + tolerance_factor * self.__std
            return array([lower]), array([upper])

        else:
            raise ValueError("The type of tolerance interval is incorrect.")



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""The silhouette coefficient to measure the quality of a clustering algorithm.

The :mod:`~gemseo.mlearning.qual_measure.silhouette` module implements the
concept of silhouette coefficient measure for machine learning algorithms.

This concept is implemented through the
:class:`.SilhouetteMeasure` class and
overloads the :meth:`!MLClusteringMeasure._compute_measure` method.

The silhouette coefficient is defined for each point as the difference
between the average distance from the point to each of the other points in its cluster
and the average distance from the point to each of the points
in the nearest cluster different from its own.

More formally,
the silhouette coefficient :math:`s_i` of a point :math:`x_i` is given by

.. math::

    a_i = \\frac{1}{|C_{k_i}| - 1} \\sum_{j\\in C_{k_i}\setminus\{i\} } \\|x_i-x_j\\|\\\\
    b_i = \\underset{\\ell=1,\\cdots,K\\atop{\\ell\\neq k_i}}{\\min}\\
        \\frac{1}{|C_\\ell|} \\sum_{j\\in C_\\ell} \\|x_i-x_j\\|\\\\
    s_i = \\frac{b_i-a_i}{\\max(b_i,a_i)}

where
:math:`k_i` is the index of the cluster to which :math:`x_i` belongs,
:math:`K` is the number of clusters,
:math:`C_k` is the set of indices of points
belonging to the cluster :math:`k` (:math:`k=1,\\cdots,K`),
and :math:`|C_k| = \\sum_{j\\in C_k} 1` is the number of points
in the cluster :math:`k`, :math:`k=1,\\cdots,K`.
"""
from __future__ import division, unicode_literals

from typing import List, Optional, Union

from numpy import ndarray
from sklearn.metrics import silhouette_score

from gemseo.core.dataset import Dataset
from gemseo.mlearning.cluster.cluster import MLPredictiveClusteringAlgo
from gemseo.mlearning.qual_measure.cluster_measure import MLPredictiveClusteringMeasure


class SilhouetteMeasure(MLPredictiveClusteringMeasure):
    """The silhouette coefficient measure for machine learning."""

    SMALLER_IS_BETTER = False

    def __init__(
        self,
        algo,  # type: MLPredictiveClusteringAlgo
    ):  # type: (...) -> None
        """
        Args:
            algo: A machine learning algorithm for clustering.
        """
        super(SilhouetteMeasure, self).__init__(algo)

    def evaluate_test(
        self,
        test_data,  # type:Dataset
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        raise NotImplementedError

    def evaluate_kfolds(
        self,
        n_folds=5,  # type: int
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        raise NotImplementedError

    def evaluate_bootstrap(
        self,
        n_replicates=100,  # type: int
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        raise NotImplementedError

    def _compute_measure(
        self,
        data,  # type: ndarray
        labels,  # type: ndarray
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        if multioutput:
            raise NotImplementedError(
                "The SilhouetteMeasure does not support the multioutput case."
            )
        measure = silhouette_score(data, labels)

        return measure



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#         Francois Gallard : refactoring for v1, May 2016
"""NLOPT library wrapper."""

from __future__ import division, unicode_literals

import logging

import nlopt
from nlopt import RoundoffLimited
from numpy import atleast_1d, atleast_2d

from gemseo.algos.opt.opt_lib import OptimizationLibrary
from gemseo.algos.stop_criteria import TerminationCriterion
from gemseo.core.function import MDOFunction

LOGGER = logging.getLogger(__name__)


class NloptRoundOffException(Exception):
    """Nlopt roundoff error."""


class Nlopt(OptimizationLibrary):
    """NLOPT optimization library interface.

    See OptimizationLibrary.
    """

    LIB_COMPUTE_GRAD = False
    STOPVAL = "stopval"
    CTOL_ABS = "ctol_abs"
    INIT_STEP = "init_step"
    SUCCESS = "NLOPT_SUCCESS: Generic success return value"
    STOPVAL_REACHED = (
        "NLOPT_STOPVAL_REACHED: Optimization stopped  "
        "because stopval (above) was reached"
    )
    FTOL_REACHED = (
        "NLOPT_FTOL_REACHED: Optimization stopped "
        "because ftol_rel or ftol_abs (above) was reached"
    )

    XTOL_REACHED = (
        "NLOPT_XTOL_REACHED Optimization stopped "
        "because xtol_rel or xtol_abs (above) was reached"
    )

    MAXEVAL_REACHED = (
        "NLOPT_MAXEVAL_REACHED: Optimization stopped "
        "because maxeval (above) was reached"
    )

    MAXTIME_REACHED = (
        "NLOPT_MAXTIME_REACHED: Optimization stopped "
        "because maxtime (above) was reached"
    )
    FAILURE = "NLOPT_FAILURE:    Generic failure code"

    INVALID_ARGS = (
        "NLOPT_INVALID_ARGS: Invalid arguments (e.g. lower "
        "bounds are bigger than upper bounds, an unknown"
        " algorithm was specified, etcetera)."
    )
    OUT_OF_MEMORY = "OUT_OF_MEMORY: Ran out of memory"
    ROUNDOFF_LIMITED = (
        "NLOPT_ROUNDOFF_LIMITED: Halted because "
        "roundoff errors limited progress. (In this "
        "case, the optimization still typically "
        "returns a useful result.)"
    )
    FORCED_STOP = (
        "NLOPT_FORCED_STOP: Halted because of a forced "
        "termination: the user called nlopt_force_stop"
        "(opt) on the optimization’s nlopt_opt"
        " object opt from the user’s objective "
        "function or constraints."
    )

    NLOPT_MESSAGES = {
        1: SUCCESS,
        2: STOPVAL_REACHED,
        3: FTOL_REACHED,
        4: XTOL_REACHED,
        5: MAXEVAL_REACHED,
        6: MAXTIME_REACHED,
        -1: FAILURE,
        -2: INVALID_ARGS,
        -3: OUT_OF_MEMORY,
        -4: ROUNDOFF_LIMITED,
        -5: FORCED_STOP,
    }

    def __init__(self):
        """Constructor."""
        super(Nlopt, self).__init__()

        nlopt_doc = "https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/"
        self.lib_dict = {
            "NLOPT_MMA": {
                self.INTERNAL_NAME: nlopt.LD_MMA,
                self.REQUIRE_GRAD: True,
                self.HANDLE_INEQ_CONS: True,
                self.HANDLE_EQ_CONS: False,
                self.DESCRIPTION: "Method of Moving Asymptotes (MMA)"
                "implemented in the NLOPT library",
                self.WEBSITE: "{}#mma-method-of-moving-asymptotes-and-ccsa".format(
                    nlopt_doc
                ),
            },
            "NLOPT_COBYLA": {
                self.INTERNAL_NAME: nlopt.LN_COBYLA,
                self.REQUIRE_GRAD: False,
                self.HANDLE_EQ_CONS: True,
                self.HANDLE_INEQ_CONS: True,
                self.DESCRIPTION: "Constrained Optimization BY Linear "
                "Approximations (COBYLA) implemented "
                "in the NLOPT library",
                self.WEBSITE: "{}".format(nlopt_doc)
                + "#cobyla-constrained-optimization-by-linear-"
                "approximations",
            },
            "NLOPT_SLSQP": {
                self.INTERNAL_NAME: nlopt.LD_SLSQP,
                self.REQUIRE_GRAD: True,
                self.HANDLE_EQ_CONS: True,
                self.HANDLE_INEQ_CONS: True,
                self.DESCRIPTION: "Sequential Least-Squares Quadratic "
                "Programming (SLSQP) implemented in "
                "the NLOPT library",
                self.WEBSITE: nlopt_doc + "#slsqp",
            },
            "NLOPT_BOBYQA": {
                self.INTERNAL_NAME: nlopt.LN_BOBYQA,
                self.REQUIRE_GRAD: False,
                self.HANDLE_EQ_CONS: False,
                self.HANDLE_INEQ_CONS: False,
                self.DESCRIPTION: "Bound Optimization BY Quadratic "
                "Approximation (BOBYQA) implemented "
                "in the NLOPT library",
                self.WEBSITE: nlopt_doc + "#bobyqa",
            },
            "NLOPT_BFGS": {
                self.INTERNAL_NAME: nlopt.LD_LBFGS,
                self.REQUIRE_GRAD: True,
                self.HANDLE_EQ_CONS: False,
                self.HANDLE_INEQ_CONS: False,
                self.DESCRIPTION: "Broyden-Fletcher-Goldfarb-Shanno method "
                "(BFGS) implemented in the NLOPT library",
                self.WEBSITE: nlopt_doc + "#low-storage-bfgs",
            },
            # Does not work on Rastrigin => banned
            #             'NLOPT_ESCH': { Does not work on Rastrigin
            #                 self.INTERNAL_NAME: nlopt.GN_ESCH,
            #                 self.REQUIRE_GRAD: False,
            #                 self.HANDLE_EQ_CONS: False,
            #                 self.HANDLE_INEQ_CONS: False},
            "NLOPT_NEWUOA": {
                self.INTERNAL_NAME: nlopt.LN_NEWUOA_BOUND,
                self.REQUIRE_GRAD: False,
                self.HANDLE_EQ_CONS: False,
                self.HANDLE_INEQ_CONS: False,
                self.DESCRIPTION: "NEWUOA + bound constraints implemented "
                "in the NLOPT library",
                self.WEBSITE: nlopt_doc + "#newuoa-bound-constraints",
            },
            # Does not work on Rastrigin => banned
            #             'NLOPT_ISRES': {
            #                 self.INTERNAL_NAME: nlopt.GN_ISRES,
            #                 self.REQUIRE_GRAD: False,
            #                 self.HANDLE_EQ_CONS: True,
            #                 self.HANDLE_INEQ_CONS: True}
        }
        for key in self.lib_dict:
            self.lib_dict[key][self.LIB] = self.__class__.__name__

    def _get_options(
        self,
        ftol_abs=1e-14,  # pylint: disable=W0221
        xtol_abs=1e-14,
        max_time=0,
        max_iter=999,
        ftol_rel=1e-8,
        xtol_rel=1e-8,
        ctol_abs=1e-6,
        stopval=None,
        normalize_design_space=True,
        eq_tolerance=1e-2,
        ineq_tolerance=1e-4,
        init_step=0.25,
        **kwargs
    ):
        r"""Sets the options

        :param max_iter: maximum number of iterations
        :type max_iter: int
        :param ftol_abs: Objective function tolerance
        :type ftol_abs: float
        :param xtol_abs: Design parameter tolerance
        :type xtol_abs: float
        :param ftol_rel: Relative objective function tolerance
        :type ftol_rel: float
        :param xtol_rel: Relative design parameter tolerance
        :type xtol_rel: float
        :param max_time: maximum runtime in seconds,
            disabled if 0 (Default value = 0)
        :type max_time: float
        :param ctol_abs: Absolute tolerance for constraints
        :type ctol_abs: float
        :param normalize_design_space: If True, scales variables in [0, 1]
        :type normalize_design_space: bool
        :param stopval: Stop when an objective value of at least stopval
            is found:
            stop minimizing when an objective value :math:`\leq` stopval is
            found,
            or stop maximizing a value :math:`\geq` stopval is found.
        :type stopval: float
        :param eq_tolerance: equality tolerance
        :type eq_tolerance: float
        :param ineq_tolerance: inequality tolerance
        :type ineq_tolerance: float
        :param kwargs: additional options
        :type kwargs: kwargs
        :param init_step: initial step size for derivavtive free algorithms
            increasing init_step will make the initial DOE in COBYLA
            wider steps in the design variables. By defaults, each variable
            is set to x0 + a perturbation that worths 0.25*(ub_i-x0_i) for i
            in xrange(len(x0))
        :type init_step: float
        """
        nds = normalize_design_space
        popts = self._process_options(
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            xtol_abs=xtol_abs,
            max_time=max_time,
            max_iter=max_iter,
            ctol_abs=ctol_abs,
            normalize_design_space=nds,
            stopval=stopval,
            eq_tolerance=eq_tolerance,
            ineq_tolerance=ineq_tolerance,
            init_step=init_step,
            **kwargs
        )

        return popts

    def __opt_objective_grad_nlopt(self, xn_vect, grad):
        """Objective function + gradient function of the optimizer for nlopt.

        :param xn_vect: normalized design vector
        :param grad: gradient array
        :returns: objective Gradients array
        :rtype: float
        """
        obj_func = self.problem.objective
        if grad.size > 0:
            grad[:] = obj_func.jac(xn_vect)
        return obj_func.func(xn_vect).real

    def __make_constraint(self, func, jac, index_cstr):
        """Builds nlopt-like constraints: no vector functions allowed The database will
        avoid multiple evaluations.

        :param func: function pointer
        :param jac: jacobian pointer
        :param index_cstr: index of the constraint
        """

        def cstr_fun_grad(xn_vect, grad):
            """Function which is given as a pointer to optimizer for constraints and
            constraints gradient if required.

            :param xn_vect: normalized design vector
            :param grad: gradient array
            """
            if self.lib_dict[self.algo_name][self.REQUIRE_GRAD]:
                if grad.size > 0:
                    cstr_jac = jac(xn_vect)
                    grad[:] = atleast_2d(cstr_jac)[
                        index_cstr,
                    ]
            return atleast_1d(func(xn_vect).real)[index_cstr]

        return cstr_fun_grad

    def __add_constraints(self, nlopt_problem, ctol=0.0):
        """Function that add all constraints (UDF+formulation) to optimization
        algorithm.

        :param nlopt_problem: optimization problem
        :rtype: unconstrainted nlopt problem
        :returns: updated nlnlopt_problem
        """
        for constraint in self.problem.constraints:
            f_type = constraint.f_type
            func = constraint.func
            jac = constraint.jac
            dim = constraint.dim
            for idim in range(dim):
                nl_fun = self.__make_constraint(func, jac, idim)
                if f_type == MDOFunction.TYPE_INEQ:
                    nlopt_problem.add_inequality_constraint(nl_fun, ctol)
                elif f_type == MDOFunction.TYPE_EQ:
                    nlopt_problem.add_equality_constraint(nl_fun, ctol)

    def __set_prob_options(self, nlopt_problem, **opt_options):
        """Settings of options for nlopt algorithms.

        :param nlopt_problem: optimization problem from nlopt
        :param opt_options: optimization options of nlopt
        """
        # ALready 0 by default
        # nlopt_problem.set_xtol_abs(0.0)
        # nlopt_problem.set_xtol_rel(0.0)
        # nlopt_problem.set_ftol_rel(0.0)
        # nlopt_problem.set_ftol_abs(0.0)
        nlopt_problem.set_maxeval(int(1.5 * opt_options[self.MAX_ITER]))  # anti-cycling
        nlopt_problem.set_maxtime(opt_options[self.MAX_TIME])
        nlopt_problem.set_initial_step(opt_options[self.INIT_STEP])
        if self.STOPVAL in opt_options:
            stopval = opt_options[self.STOPVAL]
            if stopval is not None:
                nlopt_problem.set_stopval(stopval)

        return nlopt_problem

    def _run(self, **options):
        """Runs the algorithm, to be overloaded by subclasses.

        :param options: the options dict for the algorithm,
            see associated JSON file
        """
        normalize_ds = options.pop(self.NORMALIZE_DESIGN_SPACE_OPTION, True)
        # Get the  bounds anx x0
        x_0, l_b, u_b = self.get_x0_and_bounds_vects(normalize_ds)

        nlopt_problem = nlopt.opt(self.internal_algo_name, x_0.shape[0])

        # Set the normalized bounds:
        nlopt_problem.set_lower_bounds(l_b.real)
        nlopt_problem.set_upper_bounds(u_b.real)

        nlopt_problem.set_min_objective(self.__opt_objective_grad_nlopt)
        if self.CTOL_ABS in options:
            ctol = options[self.CTOL_ABS]
        self.__add_constraints(nlopt_problem, ctol)
        nlopt_problem = self.__set_prob_options(nlopt_problem, **options)
        try:
            nlopt_problem.optimize(x_0.real)
        except (RoundoffLimited, RuntimeError) as err:
            LOGGER.error(
                "NLOPT run failed : %s, %s",
                str(err.args[0]),
                str(err.__class__.__name__),
            )
            raise TerminationCriterion()
        # status = nlopt_problem.last_optimize_result()
        message = self.NLOPT_MESSAGES[nlopt_problem.last_optimize_result()]
        status = nlopt_problem.last_optimize_result()
        return self.get_optimum_from_database(message, status)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Analytic MDODiscipline based on symbolic expressions
****************************************************
"""
from __future__ import division, unicode_literals

import logging

from numpy import array, float64, heaviside, zeros
from six import string_types
from sympy import Expr, lambdify, symbols
from sympy.parsing.sympy_parser import parse_expr

from gemseo.core.discipline import MDODiscipline
from gemseo.utils.py23_compat import PY2

LOGGER = logging.getLogger(__name__)


class AnalyticDiscipline(MDODiscipline):
    """Discipline based on analytic expressions list, using the symbolic calculation
    SymPy engine.

    Automatically differentiates the expressions to obtain
    the Jacobian matrices.

    See also
    --------
    gemseo.core.discipline.MDODiscipline : abstract class defining
        the key concept of discipline
    """

    def __init__(self, name=None, expressions_dict=None, fast_evaluation=True):
        """Constructor.

        :param name: name of the discipline.
        :type name: str
        :param expressions_dict: dictionary of outputs and their expressions
            for instance : { 'y_1':'2*x**2', 'y_2':'4*x**2+5+z**3'}
            will create a discipline with outputs y_1, y_2
            and inputs x, and z.
            Expressions may be passed as strings or SymPy expressions.
        :type expressions_dict: dict(str or sympy.Expr)
        :param fast_evaluation: if True then sympy.lambdify is applied to the
            expressions to accelerate their numerical evaluation (requires NumPy),
            otherwise the expressions are evaluated with sympy.Expr.evalf.
        :type fast_evaluation: bool
        """
        super(AnalyticDiscipline, self).__init__(name)
        if not expressions_dict:
            raise ValueError("expressions_dict is a mandatory argument")
        self.expressions_dict = expressions_dict
        self.expr_symbols_dict = {}
        self.input_names = []
        self._sympy_exprs = {}
        self._sympy_funcs = {}
        self._sympy_jac_exprs = {}
        self._sympy_jac_funcs = {}
        self._fast_evaluation = fast_evaluation
        self._init_expressions()
        self._init_grammars()
        self._init_default_inputs()
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    def _init_grammars(self):
        """Initialize the |g| grammars from the expressions dictionary."""
        zero = zeros(2)
        in_dict = {k: zero for k in self.input_names}
        self.input_grammar.initialize_from_base_dict(in_dict)
        out_dict = {k: zero for k in self.expressions_dict}
        self.output_grammar.initialize_from_base_dict(out_dict)

    def _init_expressions(self):
        """Parse the expressions (get SymPy expressions from string expressions) and
        their Jacobian expressions."""
        input_symbols = []
        for out_var, expr in self.expressions_dict.items():
            if isinstance(expr, Expr):
                parsed = expr
            elif isinstance(expr, string_types):
                parsed = parse_expr(expr)
            else:
                raise TypeError("Expression must be a SymPy expression or a string")
            args = list(parsed.free_symbols)
            self._sympy_exprs[out_var] = parsed
            input_symbols += args
            args_names = [arg.name for arg in args]
            self.expr_symbols_dict[out_var] = args_names
            self._sympy_jac_exprs[out_var] = {}
            for arg in args_names:
                jac_expr = parsed.diff(arg)
                self._sympy_jac_exprs[out_var][arg] = jac_expr

        self.input_names = sorted([symb.name for symb in set(input_symbols)])

        if self._fast_evaluation:
            self._lambdify_expressions()

    def _lambdify_expressions(self):
        """Lambdify the SymPy expressions."""
        numpy_str = "numpy"
        if PY2:
            numpy_str = numpy_str.encode("ascii")
        modules = [numpy_str, {"Heaviside": lambda x: heaviside(x, 1)}]
        for out_var, expr in self._sympy_exprs.items():
            args_names = self.expr_symbols_dict[out_var]
            args = symbols(args_names)
            self._sympy_funcs[out_var] = lambdify(args, expr)
            self._sympy_jac_funcs[out_var] = dict()
            for in_var in args_names:
                jac_expr = self._sympy_jac_exprs[out_var][in_var]
                self._sympy_jac_funcs[out_var][in_var] = lambdify(
                    args, jac_expr, modules
                )

    def _init_default_inputs(self):
        """Initialize the default inputs of the discipline with zeros."""
        zeros_array = zeros(1)
        self.default_inputs = {k: zeros_array for k in self.input_names}

    def _run(self):
        """Run the discipline."""
        outputs = {}
        # Do not pass useless tokens to the expr, this may
        # fail when tokens contains dots, or slow down the process
        filtered_inputs = {key: float(val.real) for key, val in self.local_data.items()}
        if self._fast_evaluation:
            for out_var, func in self._sympy_funcs.items():
                args_names = self.expr_symbols_dict[out_var]
                args = (filtered_inputs[name] for name in args_names)
                out_val = func(*args)
                outputs[out_var] = array([out_val], dtype=float64)
        else:
            for out_var, expr in self._sympy_exprs.items():
                try:
                    out_val = expr.evalf(subs=filtered_inputs)
                    outputs[out_var] = array([out_val], dtype=float64)
                except TypeError:
                    LOGGER.error("Failed to evaluate expression : %s", str(expr))
                    LOGGER.error("With inputs : %s", str(self.local_data))
                    raise
        self.store_local_data(**outputs)

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Compute the Jacobian.

        :param inputs: names of the differentiation inputs (Default value = None)
        :type inputs: list(str)
        :param outputs: names of the outputs to be differentiated (Default value = None)
        :type outputs: list(str)
        """
        # otherwise there may be missing terms
        # if some formula have no dependency
        self._init_jacobian(inputs, outputs, with_zeros=True)
        filtered_inputs = {key: float(val.real) for key, val in self.local_data.items()}
        if self._fast_evaluation:
            for out_var, grad in self._sympy_jac_funcs.items():
                args_names = self.expr_symbols_dict[out_var]
                for in_var, func in grad.items():
                    args = (filtered_inputs[name] for name in args_names)
                    jac_val = func(*args)
                    self.jac[out_var][in_var] = array([[jac_val]], dtype=float64)
        else:
            for out_var, expr in self._sympy_exprs.items():
                for arg in expr.free_symbols:
                    j_expr = self._sympy_jac_exprs[out_var][arg.name]
                    jac_val = j_expr.evalf(subs=filtered_inputs)
                    self.jac[out_var][arg.name] = array([[jac_val]], dtype=float64)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Data conversion between discipline data check and _run()
********************************************************
"""
from __future__ import division, unicode_literals

import logging

from numpy import array, complex128

LOGGER = logging.getLogger(__name__)


class DataProcessor(object):
    """Abstract class for pre and post processing data of MDODisciplines.

    Executes a pre processing of input data after they are checked by
    MDODiscipline.check_data, and before the _run method of the discipline
    is called.
    Similarly, the post processing method is executed after the _run method
    and before the output data is checked by |g|

    This is usefull to cast the data types, since the |g| kernel only handles
    numpy arrays, and interfaces to disciplinary tools, workflow engines or
    environment can require different types
    """

    def pre_process_data(self, data):
        """Executes a pre processing of input data after they are checked by
        MDODiscipline.check_data, and before the _run method of the discipline is
        called.

        :param data: the input data to process
        :returns: the processed input data
        """
        raise NotImplementedError()

    def post_process_data(self, data):
        """Executes a post processing of discipline output data after the _run method of
        the discipline, before they are checked by  MDODiscipline.check_output_data,

        :param data: the output data to process
        :returns: the processed output data
        """
        raise NotImplementedError()


class FloatDataProcessor(DataProcessor):
    """A data preprocessor that converts all gemseo scalar input data to floats, and
    converts all discipline output data to numpy arrays."""

    def pre_process_data(self, data):
        """Executes a pre processing of input data after they are checked by
        MDODiscipline.check_data, and before the _run method of the discipline is
        called.

        :param data: the input data to process
        :returns: the processed input data
        """
        processed_data = data.copy()
        for key, val in data.items():
            if len(val) == 1:
                processed_data[key] = float(val[0])
            else:
                processed_data[key] = [float(val_i) for val_i in val]
        return processed_data

    def post_process_data(self, data):
        """Executes a post processing of discipline output data after the _run method of
        the discipline, before they are checked by  MDODiscipline.check_output_data,

        :param data: the output data to process
        :returns: the processed output data
        """
        processed_data = data.copy()
        for key, val in data.items():
            if not hasattr(val, "__len__"):
                processed_data[key] = array([val])
            else:
                processed_data[key] = array(val)
        return processed_data


class ComplexDataProcessor(DataProcessor):
    """A data preprocessor that converts all gemseo complex arrays input data to floats
    arrays, and converts all discipline output data to numpy complex arrays."""

    def pre_process_data(self, data):
        """Executes a pre processing of input data after they are checked by
        MDODiscipline.check_data, and before the _run method of the discipline is
        called.

        :param data: the input data to process
        :returns: the processed input data
        """
        processed_data = data.copy()
        for key, val in data.items():
            processed_data[key] = array(val.real)
        return processed_data

    def post_process_data(self, data):
        """Executes a post processing of discipline output data after the _run method of
        the discipline, before they are checked by  MDODiscipline.check_output_data,

        :param data: the output data to process
        :returns: the processed output data
        """
        processed_data = data.copy()
        for key, val in data.items():
            processed_data[key] = array(val, dtype=complex128)
        return processed_data


class NameMapping(DataProcessor):
    """A data preprocessor that maps process level data names to local discipline data
    names."""

    def __init__(self, mapping):
        """Construcor.

        :param mapping: dictionary of names mapping
            data structure is {global_key:local_key}
            the global_key must be consistent with the discipline
            Grammar. The local key is the data provided to the _run
            method
        """
        super(NameMapping, self).__init__()
        self.mapping = mapping
        self.reverse_mapping = {
            local_key: global_key for global_key, local_key in mapping.items()
        }

    def pre_process_data(self, data):
        """Executes a pre processing of input data after they are checked by
        MDODiscipline.check_data, and before the _run method of the discipline is
        called.

        :param data: the input data to process
        :returns: the processed input data
        """
        mapping = self.mapping
        processed_data = {
            mapping[global_key]: value for global_key, value in data.items()
        }
        return processed_data

    def post_process_data(self, data):
        """Executes a post processing of discipline output data after the _run method of
        the discipline, before they are checked by  MDODiscipline.check_output_data,

        :param data: the output data to process
        :returns: the processed output data
        """
        reverse_mapping = self.reverse_mapping
        processed_data = {
            reverse_mapping[local_key]: value for local_key, value in data.items()
        }
        return processed_data



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
PyDOE algorithms wrapper
************************
"""
from __future__ import division, unicode_literals

import logging

from numpy.random import RandomState
from numpy.random import seed as set_seed

from gemseo.algos.doe.doe_lib import DOELibrary
from gemseo.utils.py23_compat import PY3

if PY3:
    import pyDOE2 as pyDOE
else:
    import pyDOE


LOGGER = logging.getLogger(__name__)


class PyDOE(DOELibrary):
    """PyDOE optimization library interface See DOELibrary."""

    # Available designs
    PYDOE_DOC = "https://pythonhosted.org/pyDOE/"
    PYDOE_LHS = "lhs"
    PYDOE_LHS_DESC = "Latin Hypercube Sampling implemented in pyDOE"
    PYDOE_LHS_WEB = PYDOE_DOC + "randomized.html#latin-hypercube"
    PYDOE_2LEVELFACT = "ff2n"
    PYDOE_2LEVELFACT_DESC = "2-Level Full-Factorial implemented in pyDOE"
    PYDOE_2LEVELFACT_WEB = PYDOE_DOC + "factorial.html#level-full-factorial"
    PYDOE_FULLFACT = "fullfact"
    PYDOE_FULLFACT_DESC = "Full-Factorial implemented in pyDOE"
    PYDOE_FULLFACT_WEB = PYDOE_DOC + "factorial.html#general-full-factorial"
    PYDOE_PBDESIGN = "pbdesign"
    PYDOE_PBDESIGN_DESC = "Plackett-Burman design implemented in pyDOE"
    PYDOE_PBDESIGN_WEB = PYDOE_DOC + "factorial.html#plackett-burman"
    PYDOE_BBDESIGN = "bbdesign"
    PYDOE_BBDESIGN_DESC = "Box-Behnken design implemented in pyDOE"
    PYDOE_BBDESIGN_WEB = PYDOE_DOC + "rsm.html#box-behnken"
    PYDOE_CCDESIGN = "ccdesign"
    PYDOE_CCDESIGN_DESC = "Central Composite implemented in pyDOE"
    PYDOE_CCDESIGN_WEB = PYDOE_DOC + "rsm.html#central-composite"
    ALGO_LIST = [
        PYDOE_FULLFACT,
        PYDOE_2LEVELFACT,
        PYDOE_PBDESIGN,
        PYDOE_BBDESIGN,
        PYDOE_CCDESIGN,
        PYDOE_LHS,
    ]
    DESC_LIST = [
        PYDOE_FULLFACT_DESC,
        PYDOE_2LEVELFACT_DESC,
        PYDOE_PBDESIGN_DESC,
        PYDOE_BBDESIGN_DESC,
        PYDOE_CCDESIGN_DESC,
        PYDOE_LHS_DESC,
    ]
    WEB_LIST = [
        PYDOE_FULLFACT_WEB,
        PYDOE_2LEVELFACT_WEB,
        PYDOE_PBDESIGN_WEB,
        PYDOE_BBDESIGN_WEB,
        PYDOE_CCDESIGN_WEB,
        PYDOE_LHS_WEB,
    ]
    CRITERION_KEYWORD = "criterion"
    ITERATION_KEYWORD = "iterations"
    ALPHA_KEYWORD = "alpha"
    FACE_KEYWORD = "face"
    CENTER_BB_KEYWORD = "center_bb"
    CENTER_CC_KEYWORD = "center_cc"

    def __init__(self):
        """Constructor."""
        super(PyDOE, self).__init__()
        for idx, algo in enumerate(self.ALGO_LIST):
            self.lib_dict[algo] = {
                DOELibrary.LIB: self.__class__.__name__,
                DOELibrary.INTERNAL_NAME: algo,
                DOELibrary.DESCRIPTION: self.DESC_LIST[idx],
                DOELibrary.WEBSITE: self.WEB_LIST[idx],
            }

        self.lib_dict["bbdesign"][DOELibrary.MIN_DIMS] = 3
        self.lib_dict["ccdesign"][DOELibrary.MIN_DIMS] = 2

    def _get_options(
        self,
        alpha="orthogonal",
        face="faced",
        criterion=None,
        iterations=5,
        eval_jac=False,
        center_bb=None,
        center_cc=None,
        n_samples=1,
        levels=None,
        n_processes=1,
        wait_time_between_samples=0.0,
        seed=1,
        max_time=0,
        **kwargs
    ):  # pylint: disable=W0221
        """Sets the options.

        :param alpha: effect the variance, either "orthogonal" or "rotatable"
        :type alpha: str
        :param face: The relation between the start points and the corner
            (factorial) points, either "circumscribed", "inscribed" or "faced"
        :type face: str
        :param criterion: Default value = None)
        :type criterion:
        :param iterations: Default value = 5)
        :type iterations:
        :param eval_jac: evaluate jacobian
        :type eval_jac: bool
        :param center_bb: number of center points for Box-Behnken design
        :type center_bb: int
        :param center_cc: 2-tuple of center points for the central
            composite design
        :type center_cc: tuple
        :param n_samples: number of samples
        :type n_samples: int
        :param levels: levels for axial, factorial and composite designs
        :type levels: array
        :param n_processes: number of processes
        :type n_processes: int
        :param wait_time_between_samples: waiting time between two samples
        :type wait_time_between_samples: float
        :param seed: seed value.
        :type seed: int
        :param max_time: maximum runtime in seconds,
            disabled if 0 (Default value = 0)
        :type max_time: float
        :param kwargs: additional arguments
        """
        if center_cc is None:
            center_cc = [4, 4]
        wtbs = wait_time_between_samples
        popts = self._process_options(
            alpha=alpha,
            face=face,
            criterion=criterion,
            iterations=iterations,
            center_cc=center_cc,
            center_bb=center_bb,
            eval_jac=eval_jac,
            n_samples=n_samples,
            n_processes=n_processes,
            levels=levels,
            wait_time_between_samples=wtbs,
            seed=seed,
            max_time=max_time,
            **kwargs
        )

        return popts

    @staticmethod
    def __translate(result):
        """Translate DOE design variables in [0,1]

        :param result: the samples
        """
        return (result + 1.0) * 0.5

    def _generate_samples(self, **options):
        """Generates the list of x samples.

        :param options: the options dict for the algorithm,
            see associated JSON file
        """
        self.seed += 1
        dim = options[self.DIMENSION]
        if self.algo_name == self.PYDOE_LHS:
            criterion = options.get(self.CRITERION_KEYWORD)
            iterations = options.get(self.ITERATION_KEYWORD)
            seed = options.get(self.SEED, self.seed)
            lhs_kwargs = {
                "samples": options["n_samples"],
                "criterion": criterion,
                "iterations": iterations,
            }
            if PY3:
                random_state = RandomState(seed)
                lhs_kwargs["random_state"] = random_state
            else:
                set_seed(seed)
            samples = pyDOE.lhs(dim, **lhs_kwargs)

        elif self.algo_name == self.PYDOE_CCDESIGN:
            samples = pyDOE.ccdesign(
                dim,
                center=options[self.CENTER_CC_KEYWORD],
                alpha=options[self.ALPHA_KEYWORD],
                face=options[self.FACE_KEYWORD],
            )
            # print options[self.CENTER_CC_KEYWORD], samples.shape
        elif self.algo_name == self.PYDOE_BBDESIGN:
            # Initialy designed for quadratic model fitting
            # center point is can be run several times to allow for a more
            # uniform estimate of the prediction variance over the
            # entire design space. Default value of center depends on dv_size
            samples = pyDOE.bbdesign(dim, center=options.get(self.CENTER_BB_KEYWORD))
        elif self.algo_name == self.PYDOE_FULLFACT:
            n_samples = options["n_samples"]
            self._display_fullfact_warning(n_samples)
            # Round before floor (int=floor)
            dims_list = [int(round((n_samples ** (1.0 / dim))))] * dim

            samples = pyDOE.fullfact(dims_list)

        elif self.algo_name == self.PYDOE_2LEVELFACT:
            samples = pyDOE.ff2n(dim)
        elif self.algo_name == self.PYDOE_PBDESIGN:
            samples = pyDOE.pbdesign(dim)

        if self.algo_name != self.PYDOE_LHS:
            samples = self.__translate(samples)

        if self.algo_name == self.PYDOE_FULLFACT:
            samples = self._rescale_samples(samples)

        return samples

    @staticmethod
    def is_algorithm_suited(algo_dict, problem):
        """Checks if the algorithm is suited to the problem according to its algo dict.

        :param algo_dict: the algorithm characteristics
        :param problem: the opt_problem to be solved
        """
        if DOELibrary.MIN_DIMS in algo_dict:
            if problem.dimension < algo_dict[DOELibrary.MIN_DIMS]:
                return False
        return True



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Class to fit a distribution from data based on OpenTURNS.

Overview
--------

The :class:`.OTDistributionFitter` class considers several samples
of an uncertain variable, fits a user-defined probability distribution
from this dataset and returns a :class:`.OTDistribution`.
It can also return a goodness-of-fit measure
associated with this distribution,
e.g. Bayesian Information Criterion, Kolmogorov test or Chi Squared test,
or select an optimal distribution among a collection according to
a criterion with a threshold.

Construction
------------

The :class:`.OTDistributionFitter` of a given uncertain variable is built
from only two arguments:

- a variable name,
- a one-dimensional numpy array.

Capabilities
------------

Fit a distribution
~~~~~~~~~~~~~~~~~~

The :meth:`.OTDistributionFitter.fit` method takes a distribution name
recognized by OpenTURNS as argument (e.g. 'Normal', 'Uniform', 'Exponential',
...) as argument and returns an :class:`.OTDistribution`
whose underlying OpenTURNS distribution is the specified one fitted
from the dataset passed to the constructor.

Measure the goodness-of-fit
~~~~~~~~~~~~~~~~~~~~~~~~~~~

The :meth:`.OTDistributionFitter.measure` method has two mandatory arguments:

- a distribution which is a either a :class:`.OTDistribution`
  or a distribution name from which :meth:`!fit` method
  builds a :class:`.OTDistribution`,
- a fitting criterion name.

.. note::

   Use the :meth:`.OTDistributionFitter.get_available_criteria` method to get
   the complete list of available criteria
   and the :meth:`.OTDistributionFitter.get_significance_tests` method
   to get the list of available criteria which are significance tests.

The :meth:`.OTDistributionFitter.measure` method can also use a level
associated with the criterion.

The :meth:`.OTDistributionFitter.measure` methods returns a goodness-of-fit
measure whose nature is either a scalar
when the criterion is not a significance test
or a tuple when the criterion is a significance test. In that case,
the first component of the tuple is a boolean indicating if the measured
distribution is acceptable to model the data and the second one is
a dictionary containing the test statistics, the p-value and
the significance level.

Select an optimal distribution
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The :meth:`.OTDistributionFitter.select` method select aims to select an
optimal distribution among a collection. It uses two mandatory arguments:

- a list of distribution, either a list of distributions names
  or a list of :class:`.OTDistribution`,
- a fitting criterion name.

The :meth:`.OTDistributionFitter.select` method can also use a level
associated with the criterion and a criterion selection:

- 'best': select the distribution minimizing (or maximizing, depending
  on the criterion) the criterion,
- 'first': Select the first distribution for which the criterion is
  greater (or lower, depending on the criterion) than the level.
"""

from __future__ import division, unicode_literals

import logging
from typing import Callable, List, Mapping, Sequence, Tuple, Union

import openturns as ots
from numpy import ndarray

from gemseo.uncertainty.distributions.openturns.distribution import OTDistribution
from gemseo.utils.py23_compat import string_types

LOGGER = logging.getLogger(__name__)

MeasureType = Union[Tuple[bool, Mapping[str, float]], float]


class OTDistributionFitter(object):
    """Fit a probabilistic distribution from a data array.

    Attributes:
        variable (str): The name of the variable.
        data (ndarray): The data array.
    """

    _AVAILABLE_DISTRIBUTIONS = {}
    for factory in ots.DistributionFactory.GetContinuousUniVariateFactories():
        factory_class_name = factory.getImplementation().getClassName()
        dist_name = factory_class_name.split("Factory")[0]
        _AVAILABLE_DISTRIBUTIONS[dist_name] = getattr(ots, factory_class_name)

    AVAILABLE_DISTRIBUTIONS = sorted(_AVAILABLE_DISTRIBUTIONS.keys())

    _AVAILABLE_FITTING_TESTS = {
        "BIC": ots.FittingTest.BIC,
        "Kolmogorov": ots.FittingTest.Kolmogorov,
        "ChiSquared": ots.FittingTest.ChiSquared,
    }
    AVAILABLE_FITTING_TESTS = sorted(_AVAILABLE_FITTING_TESTS.keys())

    SIGNIFICANCE_TESTS = ["Kolmogorov", "ChiSquared"]

    _CRITERIA_TO_MINIMIZE = ["BIC"]

    def __init__(
        self,
        variable,  # type: str
        data,  # type: ndarray
    ):  # noqa: D205,D212,D415
        # type: (...) -> None
        """
        Args:
            variable: The name of the variable.
            data: A data array.
        """
        self.variable = variable
        try:
            isinstance(data, ndarray)
            self.data = ots.Sample(data.reshape((-1, 1)))
        except AttributeError:
            raise TypeError("data must be a numpy array")

    def _get_factory(
        self,
        distribution,  # type: str
    ):
        # type: (...) -> ots.DistributionFactory
        """Get the distribution factory.

        Args:
            distribution: The name of the distribution.

        Returns:
            An OpenTURNS distribution factory.
        """
        try:
            distribution_factory = self._AVAILABLE_DISTRIBUTIONS[distribution]
        except KeyError:
            distributions = ", ".join(list(self._AVAILABLE_DISTRIBUTIONS.keys()))
            raise ValueError(
                "{} is not a name of "
                "distribution available for fitting. "
                "Available ones are: {}.".format(distribution, distributions)
            )
        return distribution_factory

    def _get_fitting_test(
        self,
        criterion,  # type: str
    ):
        # type: (...) -> Callable
        """Get the fitting test.

        Args:
            criterion: The name of a fitting criterion.

        Returns:
            The OpenTURNS fitting test corresponding to the provided name.
        """
        try:
            fitting_test = self._AVAILABLE_FITTING_TESTS[criterion]
        except KeyError:
            tests = ", ".join(list(self._AVAILABLE_FITTING_TESTS.keys()))
            raise ValueError(
                "%s is not a name of fitting test; "
                "Available ones are: %s" % (criterion, tests)
            )
        return fitting_test

    def fit(
        self,
        distribution,  # type: str
    ):
        # type: (...) -> OTDistribution
        """Fit a distribution.

        Args:
            distribution: The name of a distribution.

        Returns:
            The distribution corresponding to the provided name.
        """
        factory = self._get_factory(distribution)
        fitted_distribution = factory().build(self.data)
        parameters = fitted_distribution.getParameter()
        distribution = OTDistribution(self.variable, distribution, parameters)
        return distribution

    def compute_measure(
        self,
        distribution,  # type: Union[OTDistribution, str]
        criterion,  # type: str
        level=0.05,  # type: float
    ):
        # type: (...) -> MeasureType
        """Measure the goodness-of-fit of a distribution to data.

        Args:
            distribution: A distribution name.
            criterion: The name of the goodness-of-fit criterion.
            level: A test level,
                i.e. the risk of committing a Type 1 error,
                that is an incorrect rejection of a true null hypothesis,
                for criteria based on test hypothesis.

        Returns:
            The goodness-of-fit measure.
        """
        if isinstance(distribution, string_types):
            distribution = self.fit(distribution)
        if distribution.dimension > 1:
            raise TypeError("A 1D distribution is required.")
        distribution = distribution.marginals[0]
        fitting_test = self._get_fitting_test(criterion)
        if criterion in self.SIGNIFICANCE_TESTS:
            result = fitting_test(self.data, distribution, level)
            details = {
                "p-value": result.getPValue(),
                "statistics": result.getStatistic(),
                "level": level,
            }
            result = (result.getBinaryQualityMeasure(), details)
        else:
            result = fitting_test(self.data, distribution)
        return result

    def select(
        self,
        distributions,  # type: Union[Sequence[str], Sequence[OTDistribution]]
        fitting_criterion,  # type: str
        level=0.05,  # type: float
        selection_criterion="best",  # type: str
    ):
        # type: (...) -> OTDistribution
        """Select the best distribution from a list of candidates.

        Args:
            distributions: The distributions.
            fitting_criterion: The name of the goodness-of-fit criterion.
            level: A test level,
                i.e. the risk of committing a Type 1 error,
                that is an incorrect rejection of a true null hypothesis,
                for criteria based on test hypothesis.
            selection_criterion:  The name of the selection criterion.
                Either 'first' or 'best'.

        Returns:
            The best distribution.
        """
        measures = []
        for index, distribution in enumerate(distributions):
            if isinstance(distribution, string_types):
                distribution = self.fit(distribution)
            measures.append(
                self.compute_measure(distribution, fitting_criterion, level)
            )
            distributions[index] = distribution
        index = self.select_from_measures(
            measures, fitting_criterion, level, selection_criterion
        )
        return distributions[index]

    @classmethod
    def select_from_measures(
        cls,
        measures,  # type: List[MeasureType]
        fitting_criterion,  # type: str
        level=0.05,  # type: float
        selection_criterion="best",  # type: str
    ):
        # type: (...) -> int
        """Select the best distribution from measures.

        Args:
            measures: The measures.
            fitting_criterion: The name of the goodness-of-fit criterion.
            level: A test level,
                i.e. the risk of committing a Type 1 error,
                that is an incorrect rejection of a true null hypothesis,
                for criteria based on test hypothesis.
            selection_criterion:  The name of the selection criterion.
                Either 'first' or 'best'.

        Returns:
            The index of the best distribution.
        """
        if fitting_criterion in cls.SIGNIFICANCE_TESTS:
            for index, _ in enumerate(measures):
                measures[index] = measures[index][1]["p-value"]
            if sum([p_value > level for p_value in measures]) == 0:
                LOGGER.warning(
                    "All criteria values are lower than the significance level %s.",
                    level,
                )
        if selection_criterion == "best" or level is None:
            index = cls.__find_opt_distribution(measures, fitting_criterion)
        else:
            index = cls.__apply_first_strategy(measures, fitting_criterion, level)
        return index

    @classmethod
    def __apply_first_strategy(
        cls,
        measures,  # type: List[float]
        fitting_criterion,  # type: str
        level=0.05,  # type: float
    ):  # type: (...) -> int
        """Select the best distribution from measures by applying the "first" strategy.

        Args:
            measures: The measures.
            fitting_criterion: The name of the goodness-of-fit criterion.
            level: A test level,
                i.e. the risk of committing a Type 1 error,
                that is an incorrect rejection of a true null hypothesis,
                for criteria based on test hypothesis.

        Returns:
            The index of the best distribution.
        """
        select = False
        index = 0
        for measure in measures:
            select = measure >= level
            if select:
                break
            index += 1
        if not select:
            index = cls.__find_opt_distribution(measures, fitting_criterion)
        return index

    @classmethod
    def __find_opt_distribution(
        cls,
        measures,  # type: List[float]
        fitting_criterion,  # type: str
    ):
        # type: (...) -> int
        """Select the best distribution from measures by applying the "best" strategy.

        Args:
            measures: The measures.
            fitting_criterion: The name of the goodness-of-fit criterion.

        Returns:
            The index of the optimum distribution.
        """
        if fitting_criterion in cls._CRITERIA_TO_MINIMIZE:
            index = measures.index(min(measures))
        else:
            index = measures.index(max(measures))
        return index

    @property
    def available_distributions(self):
        # type: (...) -> List[str]
        """The available distributions."""
        return sorted(self._AVAILABLE_DISTRIBUTIONS.keys())

    @property
    def available_criteria(self):
        # type: (...) -> List[str]
        """The available goodness-of-fit criteria."""
        return sorted(self._AVAILABLE_FITTING_TESTS.keys())

    @property
    def available_significance_tests(self):
        # type: (...) -> List[str]
        """The significance tests."""
        return sorted(self.SIGNIFICANCE_TESTS)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Generator of the JSON file defining the coupling structure used by the N2 chart."""
from __future__ import unicode_literals

import json
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from jinja2 import Template

if TYPE_CHECKING:
    from gemseo.core.coupling_structure import DependencyGraph

from gemseo.core.discipline import MDODiscipline


class N2JSON(object):
    """The JSON structure to be used by the D3.js-based N2 chart."""

    _DEFAULT_GROUP_TEMPLATE = "Group {}"
    _DEFAULT_WEAKLY_COUPLED_DISCIPLINES = "Weakly coupled disciplines"

    def __init__(
        self,
        graph,  # type: DependencyGraph
    ):  # type: (...) -> None
        """
        Args:
            graph: The dependency graph.
        """
        self._graph = graph
        self.__disciplines = list(graph.disciplines)

        data = {}

        self.__disciplines_names = self._get_disciplines_names()

        couplings = self._graph.get_disciplines_couplings()

        groups, n_groups, data["children"] = self._compute_groups(
            self.__disciplines_names
        )

        variables_sizes = self._compute_variables_sizes()

        data["nodes"], data["groups"] = self._create_nodes(
            groups,
            variables_sizes,
            self.__disciplines_names,
            n_groups,
            data["children"],
        )

        data["links"] = self._create_links(
            couplings,
            len(data["nodes"]),
            variables_sizes,
            self.__disciplines_names,
            n_groups,
        )

        data["disciplines"] = self._create_groups_menu_html(
            self.__disciplines_names, data["children"], data["groups"]
        )

        self.__json = json.dumps(data, sort_keys=True)

    def __str__(self):
        return self.__json

    @staticmethod
    def _create_variables_html(
        names,  # type: Iterable[str]
        variables_sizes=None,  # type: Optional[Mapping[str,int]]
    ):  # type: (...) -> str
        """Generate the HTML representation of variables from their names and sizes.

        Args:
            names: The names of the variables.
            variables_sizes: The sizes of the variables.
                If None, display only the names.

        Return:
            The HTML representation of the sorted variables.
        """
        variables = [
            {
                "name": name,
                "size": None
                if variables_sizes is None
                else variables_sizes.get(name, 1),
            }
            for name in sorted(names)
        ]

        return Template(
            "<div align='center'>"
            "    <table>"
            "    {%- for variable in variables %}"
            "        <tr>"
            "            <td>{{ variable.name }}</td>"
            "            {%- if variable.size is not none -%}"
            "            <td>({{ variable.size }})</td>"
            "            {%- endif %}"
            "        </tr>"
            "    {%- endfor %}"
            "    </table>"
            "</div>"
        ).render(variables=variables)

    @classmethod
    def _create_coupling_html(
        cls,
        source,  # type: str
        destination,  # type: str
        coupling_names,  # type: Iterable[str]
        variables_sizes,  # type: Mapping[str,int]
    ):  # type: (...) -> str
        """Generate the HTML representation of a bi-disciplinary coupling.

        Args:
            source: The name of the source discipline.
            destination: The name of the destination discipline.
            coupling_names: The names of the coupling variables.
            variables_sizes: The sizes of the variables.

        Returns:
            The HTML block describing this bi-disciplinary coupling.
        """
        return Template(
            (
                "<h2 style='text-align: center;'>"
                "<span style='color: gray;'>Coupling</span></br>"
                "<span style='color: gray;'>from</span> {{ source }}<br/>"
                "<span style='color: gray;'>to</span> {{ destination }}</h2>"
                "{{ coupling_variables }}"
            )
        ).render(
            source=source,
            destination=destination,
            coupling_variables=cls._create_variables_html(
                coupling_names, variables_sizes
            ),
        )

    @classmethod
    def _create_discipline_html(
        cls,
        discipline,  # type: MDODiscipline
        variables_sizes,  # type: Mapping[str,int]
    ):  # type: (...) -> str
        """Generate the HTML representation of a discipline.

        Args:
            discipline: The discipline.
            variables_sizes: The sizes of the variables.

        Returns:
            The HTML block describing the discipline.
        """
        html_inputs_names = cls._create_variables_html(
            discipline.get_input_data_names(), variables_sizes
        )
        html_outputs_names = cls._create_variables_html(
            discipline.get_output_data_names(), variables_sizes
        )
        return Template(
            (
                "<h2 style='text-align: center;'>{{ discipline }}</h2>"
                "<h3 style='text-align: center;'>Inputs</h3>"
                "{{ inputs }}"
                "<h3 style='text-align: center;'>Outputs</h3>"
                "{{ outputs }}"
            )
        ).render(
            discipline=discipline.name,
            inputs=html_inputs_names,
            outputs=html_outputs_names,
        )

    @classmethod
    def _create_group_html(
        cls,
        group,  # type: int
        disciplines,  # type: Sequence[str]
        n_groups,  # type: int
        children,  # type: Sequence[Sequence[int]]
    ):  # type: (...) -> str
        """Generate the HTML representation of a group of disciplines.

        Args:
            group: The index of the group of disciplines.
            disciplines: The names of the disciplines.
            n_groups: The number of groups.
            children: The indices of the disciplines for the different groups.

        Returns:
            The HTML block describing the group of disciplines.
        """
        disciplines = [disciplines[child - n_groups] for child in children[group]]
        return Template(
            "<h2 style='text-align: center;'>{{ group }}</h2>{{ disciplines }}"
        ).render(
            group=cls._DEFAULT_GROUP_TEMPLATE.format(group),
            disciplines=cls._create_variables_html(disciplines),
        )

    @staticmethod
    def _create_groups_menu_html(
        disciplines,  # type: Sequence[str]
        children,  # type: Sequence[Sequence[int]]
        groups,  # type: Sequence[str]
    ):  # type: (...) -> str
        """Generate the HTML representation of the right menu related to the groups.

        Args:
            disciplines: The names of the disciplines.
            children: The indices of the disciplines for the different groups.
            groups: The names of the groups.

        Returns:
            The HTML block used to collapse, expand and visualize groups.
        """
        data = []
        n_groups = len(groups)
        for group_index, disciplines_indices in enumerate(children):
            disciplines_names = [
                disciplines[discipline_index - n_groups]
                for discipline_index in disciplines_indices
            ]
            data.append(
                {
                    "disciplines": sorted(disciplines_names),
                    "group_index": group_index,
                    "group_name": groups[group_index],
                }
            )
        return Template(
            "<h2>Group the disciplines</h2>"
            "<div style='overflow:scroll; height:39em;'>"
            "    <ul id='myUL'>"
            "        {%- for datum in data %}"
            "        <li>"
            "            <div class='caret'>"
            "                <span id='group_name_{{ datum.group_index }}' contenteditable='true' class='group' onblur='change_group_name(this,{{ datum.group_index }});'>{{ datum.group_name }}</span>"  # noqa: B950
            "                {%- if datum.group_index != 0 %}"
            "                <input type='checkbox' onclick='expand_collapse_group({{ datum.group_index }},svg)'/>"  # noqa: B950
            "                {%- endif %}"
            "            </div>"
            "            <ul class='nested'>"
            "               {%- for discipline in datum.disciplines %}"
            "                <li>{{ discipline }}</li>"
            "                {%- endfor %}"
            "            </ul>"
            "        </li>"
            "        {%- endfor %}"
            "    </ul>"
            "</div>"
        ).render(data=data)

    def _create_links(
        self,
        couplings,  # type: Iterable[Tuple[MDODiscipline,MDODiscipline,Sequence[str]]]
        n_nodes,  # type: int
        variables_sizes,  # type: Mapping[str,int]
        disciplines,  # type: Sequence[str]
        n_groups,  # type: int
    ):  # type: (...) -> List[Dict[str,Union[int,str]]]
        """Create the links.

        Args:
            couplings: The couplings.
            n_nodes: The number of nodes.
            variables_sizes: The sizes of the variables.
            disciplines: The names of the disciplines.
            n_groups: The number of groups.

        Returns:
            The existing links between disciplines,
            defined by a source discipline
            a target discipline,
            a value quantifying the degree of this relationship,
            and an HTML representation.
        """

        links = []
        for link in couplings:
            source, target, variables = link
            if variables:
                source_index = n_groups + disciplines.index(source.name)
                target_index = n_groups + disciplines.index(target.name)
                links.append(
                    {
                        "source": source_index,
                        "target": target_index,
                        "value": len(variables),
                        "description": self._create_coupling_html(
                            source.name, target.name, variables, variables_sizes
                        ),
                    }
                )

        for index in range(n_nodes):
            links.append(
                {
                    "source": index,
                    "target": index,
                    "value": 1,
                    "description": "",
                }
            )

        return links

    def _create_nodes(
        self,
        group,  # type: Mapping[str,int],
        variables_sizes,  # type: Mapping[str,int],
        disciplines,  # type: Sequence[str]
        n_groups,  # type: int
        children,  # type: Sequence[Sequence[int]]
    ):  # type: (...) -> Tuple[List[Dict[str,Union[int,str,bool]]],List[str]]
        """Create the nodes representing either a discipline or a disciplines group.

        Args:
            group: The indices of the groups to which the disciplines belong.
            variables_sizes: The sizes of the variables.
            disciplines: The names of the disciplines.
            n_groups: The number of groups.
            children: The indices of the disciplines for the different groups.

        Returns:
            The existing nodes,
            defined by a name,
            a group index,
            an HTML representation
            a target discipline,
            a value quantifying the degree of this relationship,
            an HTML representation,
            and whether the node represents a group.
        """
        disciplines_nodes = [
            {
                "name": discipline.name,
                "group": group[discipline.name],
                "description": self._create_discipline_html(
                    discipline, variables_sizes
                ),
                "is_group": False,
            }
            for discipline in self.__disciplines
        ]

        groups_nodes = [
            {
                "name": self._DEFAULT_GROUP_TEMPLATE.format(group_index),
                "group": group_index,
                "description": self._create_group_html(
                    group_index, disciplines, n_groups, children
                ),
                "is_group": True,
            }
            for group_index in range(n_groups)
        ]

        groups_names = [
            self._DEFAULT_GROUP_TEMPLATE.format(group_index)
            for group_index in range(1, n_groups)
        ]
        groups_names = [self._DEFAULT_WEAKLY_COUPLED_DISCIPLINES] + groups_names

        return groups_nodes + disciplines_nodes, groups_names

    def _compute_groups(
        self,
        disciplines,  # type: Sequence[str]
    ):  # type: (...) -> Tuple[Dict[str,int],int,List[List[int]]]
        """Compute the groups and the children.

        Args:
            disciplines: The names of the disciplines

        Returns:
            The groups to which the disciplines belong,
            the number of groups,
            and the indices of the disciplines for the different groups.
        """
        children = [[]]
        n_groups = 1
        groups = {}
        for parallel_tasks in self._graph.get_execution_sequence():
            for components in parallel_tasks:
                if len(components) > 1:
                    n_groups += 1
                    children.append([])
                    for component in components:
                        index = disciplines.index(component.name)
                        children[-1].append(index)
                        groups[component.name] = n_groups - 1
                else:
                    index = disciplines.index(components[0].name)
                    groups[components[0].name] = 0
                    children[0].append(index)
        indices = []
        new_children = []
        index = 0
        for group in children:
            new_children.append([])
            for child in group:
                new_children[-1].append(index)
                index += 1
                indices.append(child)
        self.__disciplines = [self.__disciplines[index] for index in indices]
        self.__disciplines_names = [
            self.__disciplines_names[index] for index in indices
        ]

        new_children = [[child + n_groups for child in group] for group in new_children]
        return groups, n_groups, new_children

    def _compute_variables_sizes(self):  # type: (...) -> Dict[str,int]
        """Compute the sizes of the coupling variables.

        Returns:
            The names of the coupling variables bound to their sizes.
        """
        variables_sizes = {}
        for discipline in self.__disciplines:
            for name in discipline.get_input_data_names():
                variables_sizes[name] = len(discipline.default_inputs.get(name, [1]))
        return variables_sizes

    def _get_disciplines_names(self):  # type: (...) -> List[str]
        """Return the names of the disciplines."""
        return [discipline.name for discipline in self.__disciplines]
